# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from IProgress import ProgressBar, Bar, Percentage
from bokeh.charts import Line
from bokeh.layouts import column
from bokeh.models import FactorRange, Range1d, LinearAxis
from bokeh.plotting import figure, show
from cameo.core.result import Result
from cameo.flux_analysis.analysis import flux_variability_analysis, FluxVariabilityResult
from cameo.flux_analysis.simulation import pfba, fba
from cobra.core.metabolite import Metabolite
from cobra.core.model import Model
from cobra.core.reaction import Reaction
from cobra.exceptions import OptimizationError, Infeasible
from numpy import array
from pandas import DataFrame

from marsi.cobra.flux_analysis.manipulation import knockout_metabolite, compete_metabolite, inhibit_metabolite
from marsi.utils import frange


__all__ = ['metabolite_knockout_fitness', 'metabolite_knockout_phenotype', 'sensitivity_analysis']


BASE_ELEMENTS = ["C", "N"]


logger = logging.getLogger(__name__)


class MetaboliteKnockoutFitness(Result):
    """
    The result holder for the fitness landscape analysis.

    See Also
    --------
    marsi.cobra.flux_analysis.analysis.metabolite_knockout_fitness

    Attributes
    ----------
    data_frame : pandas.DataFrame
        A DataFrame with the result.


    """
    def __init__(self, fitness_data_frame, *args, **kwargs):
        super(MetaboliteKnockoutFitness, self).__init__(*args, **kwargs)
        assert isinstance(fitness_data_frame, DataFrame)
        self._data_frame = fitness_data_frame

    @property
    def data_frame(self):
        return DataFrame(self._data_frame)

    def plot(self, grid=None, width=None, height=None, title=None, *args, **kwargs):
        """
        """
        data = self._data_frame.sort_values('fitness')
        data['x'] = data.index
        show(Line(data, 'x', 'fitness', title=title, plot_width=width, plot_height=height))

    def _repr_html_(self):
        return self.plot(height=500, width=12 * len(self._data_frame), title="Metabolite knockout fitness landscape")


def metabolite_knockout_fitness(model, simulation_method=pfba, compartments=None, elements=BASE_ELEMENTS,
                                objective=None, ndecimals=6, progress=False, ncarbons=2, **simulation_kwargs):
    """
    Calculate the landscape of fitness for each metabolite knockout in the model.

    Parameters
    ----------
    model : cameo.core.SolverBasedModel
        A constraint-based model.
    simulation_method : cameo.flux_analysis.simulation.fba
        A method to simulate the knockouts (e.g. cameo.flux_analysis.simulation.moma)
    compartments : list
        The compartments to consider (e.g. ["c", "g", "r"] for cytosol, golgi aparatus and endoplasmatic reticulum)
    elements : list
        Atomic elements to add to the result.
    objective : str, cameo.core.Reaction, other
        A valid objective for the model.
    ndecimals : int
        Number of decimals to use as precision.
    progress : bool
        Report progress.
    ncarbons : int
        Minimum number of carbons to consider.
    simulation_kwargs : dict
        Arguments for `simulation_method

    Returns
    -------
    MetaboliteKnockoutFitness
        The fitness landscape.
    """
    assert isinstance(model, Model)
    fitness = DataFrame(columns=["fitness"] + list(elements))

    if compartments is None:
        compartments = list(model.compartments.keys())

    if progress:
        iterator = ProgressBar(maxval=len(model.metabolites), widgets=[Bar(), Percentage()])
    else:
        iterator = iter
    for met in iterator(model.metabolites):
        if met.compartment in compartments and met.elements.get("C", 0) > ncarbons:
            with model:
                knockout_metabolite(model, met, allow_accumulation=True, ignore_transport=True)
                try:
                    solution = simulation_method(model, objective=objective, **simulation_kwargs)
                    fitness.loc[met.id] = [round(solution[objective], ndecimals)] + \
                                          [met.elements.get(el, 0) for el in elements]
                except OptimizationError:
                    fitness.loc[met.id] = [.0] + [met.elements.get(el, 0) for el in elements]

    return MetaboliteKnockoutFitness(fitness)


class MetaboliteKnockoutPhenotypeResult(MetaboliteKnockoutFitness):
    def __init__(self, phenotype_data_frame, *args, **kwargs):
        super(MetaboliteKnockoutPhenotypeResult, self).__init__(phenotype_data_frame, *args, **kwargs)

    def __getitem__(self, item):
        phenotype = {}
        if isinstance(item, (int, slice)):
            fva = self._data_frame['fva'].iloc[item]
        elif isinstance(item, str):
            fva = self._data_frame['fva'].loc[item]
        elif isinstance(item, Metabolite):
            fva = self._data_frame['fva'].loc[item.id]
        else:
            raise ValueError("%s is not a valid search retrieval")

        for reaction_id, row in fva.data_frame.iterrows():
            if row['upper_bound'] == 0 and row['lower_bound'] == 0:
                continue
            phenotype[reaction_id] = (row['upper_bound'], row['lower_bound'])

        return phenotype

    def _phenotype_plot(self, index, factors):
        plot = figure(title=index, y_range=FactorRange(factors))
        phenotype = self[index]

        y0 = [phenotype.get(f, [0, 0])[0] for f in factors]
        y1 = [phenotype.get(f, [0, 0])[1] for f in factors]

        x0 = [(i - .5) for i in range(len(factors))]
        x1 = [(i + .5) for i in range(len(factors))]

        plot.quad(x0, x1, y0, y1)

        return plot

    def plot(self, indexes=None, grid=None, width=None, height=None, title=None, conditions=None, *args, **kwargs):
        data = self.data_frame
        if conditions:
            data = data.query(conditions)
        if indexes:
            data = data.loc[indexes]

        factors = list(set(sum(data.phenotype.apply(lambda p: list(p.keys())), [])))
        plots = []

        for index in data.index:
            plots.append(self._phenotype_plot(index, factors))
        show(column(plots))


def metabolite_knockout_phenotype(model, compartments=None, objective=None, ndecimals=6, elements=BASE_ELEMENTS,
                                  progress=False, ncarbons=2):
    assert isinstance(model, Model)
    phenotype = DataFrame(columns=['fitness', 'fva'] + elements)
    exchanges = model.exchanges

    if progress:
        iterator = ProgressBar(maxval=len(model.metabolites), widgets=[Bar(), Percentage()])
    else:
        iterator = iter

    for met in iterator(model.metabolites):
        if met.compartment in compartments and met.elements.get("C", 0) > ncarbons:
            with model:
                knockout_metabolite(model, met, allow_accumulation=True, ignore_transport=True)
                fitness = fba(model, objective=objective)
                fva = flux_variability_analysis(model, reactions=exchanges, fraction_of_optimum=1)
                fva = FluxVariabilityResult(fva.data_frame.apply(round, args=(ndecimals,)))
                phenotype.loc[met.id] = [fitness, fva] + [met.elements.get(el, 0) for el in elements]

    return MetaboliteKnockoutPhenotypeResult(phenotype)


class SensitivityAnalysisResult(Result):
    def __init__(self, species_id, exchange_fluxes, steps, is_essential, biomass_fluxes=None,
                 biomass=None, variables=None, variables_fluxes=None, *args, **kwargs):
        super(SensitivityAnalysisResult, self).__init__(*args, **kwargs)
        self._is_essential = is_essential
        self._species_id = species_id
        self._exchange_fluxes = exchange_fluxes
        self._steps = steps
        self._biomass_fluxes = biomass_fluxes
        self._biomass = biomass
        self._variables = variables
        self._variables_fluxes = variables_fluxes

    @property
    def data_frame(self):

        data = [self._steps, self._exchange_fluxes]
        index = ["fraction", self._species_id]

        if self._biomass is not None:
            data.append(self._biomass_fluxes)
            index.append(self._biomass.id)

        if self._variables is not None and len(self._variables) > 0:
            for i, variable in enumerate(self._variables):
                data.append(self._variables_fluxes[:, i].tolist())
                if isinstance(variable, Reaction):
                    variable = variable.id
                index.append(variable)

        return DataFrame(data, index=index).T

    def plot(self, grid=None, width=None, height=None, *args, **kwargs):
        x_label = "Competition Level" if self._is_essential else 'Inhibition level'
        fig = figure(plot_width=width, plot_height=height, title=self._species_id,
                     x_axis_label=x_label, y_axis_label="Accumulation Level (mmol/gDW)",
                     toolbar_sticky=False)

        data = self.data_frame
        fig.extra_y_ranges = {}
        if self._biomass is not None:
            fig.extra_y_ranges["growth_rate"] = Range1d(start=0, end=data[self._biomass.id].max())
            fig.add_layout(LinearAxis(y_range_name="growth_rate", axis_label="Growth Rate [h^-1])"), 'right')
        if self._variables is not None and len(self._variables) > 0:
            fig.extra_y_ranges["flux"] = Range1d(start=min(data[data.columns[3:]].min()),
                                                 end=max(data[data.columns[3:]].max()))
            fig.add_layout(LinearAxis(y_range_name="flux", axis_label="Flux [mmol h^-1 gDW^-1])"), 'right')
        if self._biomass is None:
            fig.line(data['fraction'].apply(lambda v: v if self._is_essential else 1 - v) * 100,
                     data[self._species_id], line_color='orange')

        else:
            fig.line(data['fraction'].apply(lambda v: v if self._is_essential else 1 - v) * 100,
                     data[self._species_id], line_color='orange')
            fig.line(data['fraction'].apply(lambda v: v if self._is_essential else 1 - v) * 100,
                     data[self._biomass.id], line_color='green',
                     y_range_name="growth_rate")

        for var in self._variables:
            if isinstance(var, Reaction):
                var = var.id
            fig.line(data['fraction'].apply(lambda v: v if self._is_essential else 1 - v) * 100,
                     data[var], y_range_name="flux", legend=var)

        show(fig)


def sensitivity_analysis(model, metabolite, biomass=None, variables=None, is_essential=False, steps=10,
                         reference_dist=None, simulation_method=fba, **simulation_kwargs):

    if reference_dist is None:
        simulation_kwargs['reference'] = simulation_kwargs.get('reference', None) or pfba(model, objective=biomass)

    if variables is None:
        variables = []

    species_id = metabolite.id[:-2]

    exchange_fluxes = []
    biomass_fluxes = []
    variables_fluxes = []
    fractions = []

    for i, fraction in enumerate(frange(0, 1.1, steps)):
        variables_fluxes.append([])
        with model:
            if is_essential:
                exchange = compete_metabolite(model, metabolite, simulation_kwargs['reference'], fraction)
            else:
                exchange = inhibit_metabolite(model, metabolite, simulation_kwargs['reference'], fraction)
            try:
                flux_dist = simulation_method(model, objective=biomass, **simulation_kwargs)

                if biomass is not None:
                    biomass_fluxes.append(flux_dist[biomass])

                for variable in variables:
                    variables_fluxes[i].append(flux_dist[variable])

                flux = flux_dist[exchange]

                exchange_fluxes.append(flux)
                fractions.append(fraction)
                logger.debug("Feasible: %s (%.3f) essential: %s flux: %.3f" %
                             (species_id, fraction, is_essential, flux))
            except Infeasible:
                logger.debug("Infeasible: %s (%.3f) essential: %s" % (species_id, fraction, is_essential))
                if biomass is not None:
                    biomass_fluxes.append(0)

                for _ in variables:
                    variables_fluxes[i].append(0)

                exchange_fluxes.append(0)
                fractions.append(fraction)

    return SensitivityAnalysisResult(species_id, exchange_fluxes, fractions, is_essential, biomass_fluxes, biomass,
                                     variables, array(variables_fluxes))



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import math
import time

import numpy as np
import rdkit
from bitarray import bitarray
from cachetools import cached, LRUCache
from rdkit import Chem
from rdkit.Chem import MCS, AllChem, MACCSkeys
from rdkit.Chem.SaltRemover import SaltRemover

from marsi.chemistry.common import monte_carlo_volume as mc_vol, inchi_key_lru_cache


lru_cache = LRUCache(maxsize=256)

periodic_table = Chem.GetPeriodicTable()
salt_remove = SaltRemover(defnData="[Li,K,Rb,Cs,Fr,Be,Mg,Ca,Sr,Ba,Ra,F,Cl,Br]")

fps = ["maccs", "morgan2", "morgan3", "morgan4", "morgan5"]


@cached(lru_cache)
def inchi_to_molecule(inchi):
    """
    Returns a molecule from a InChI string.

    Parameters
    ----------
    inchi : str
        A valid string.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        A molecule.
    """
    mol = Chem.MolFromInchi(inchi)
    mol = salt_remove.StripMol(mol, dontRemoveEverything=True)
    Chem.Kekulize(mol)
    mol = Chem.AddHs(mol)

    return mol


def mol_to_molecule(file_or_molecule_desc, from_file=True):
    """
    Returns a molecule from a MOL file.

    Parameters
    ----------
    file_or_molecule_desc : str
        A valid MOL file path or a valid MOL string.
    from_file : bool
        If True tries to read the molecule from a file.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        A molecule.
    """
    if from_file:
        mol = Chem.MolFromMolFile(file_or_molecule_desc)
    else:
        mol = Chem.MolFromMolBlock(str(file_or_molecule_desc))
    mol = salt_remove.StripMol(mol, dontRemoveEverything=True)
    Chem.Kekulize(mol)
    mol = Chem.AddHs(mol)
    return mol


def sdf_to_molecule(file_or_molecule_desc, from_file=True):
    """
    Returns a molecule from a SDF file.

    Parameters
    ----------
    file_or_molecule_desc : str
        A valid sdf file path or a valid SDF string.
    from_file : bool
        If True tries to read the molecule from a file.

    Returns
    -------
    rdkit.Chem.rdchem.Mol
        A molecule.
    """

    if from_file:
        supplier = Chem.SDMolSupplier(file_or_molecule_desc)
    else:
        supplier = Chem.SDMolSupplier()
        supplier.SetData(str(file_or_molecule_desc), strictParsing=False)
    mol = next(supplier)
    mol = salt_remove.StripMol(mol, dontRemoveEverything=True)
    Chem.Kekulize(mol)
    mol = Chem.AddHs(mol)
    return mol


@cached(inchi_key_lru_cache)
def inchi_to_inchi_key(inchi):
    """
    Makes an InChI Key from a InChI string.

    Parameters
    ----------
    inchi : str
        A valid InChI string.

    Returns
    -------
    str
        A InChI key.
    """
    return Chem.InchiToInchiKey(inchi)


def mol_to_inchi_key(mol):
    """
    Makes an InChI Key from a Molecule.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        A molecule.

    Returns
    -------
    str
        A InChI key.
    """
    return inchi_to_inchi_key(mol_to_inchi(mol))


def mol_to_inchi(mol):
    """
    Makes an InChI from a Molecule.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        A molecule.

    Returns
    -------
    str
        A InChI.
    """
    return Chem.MolToInchi(mol)


def fingerprint(molecule, fpformat='maccs'):
    """
    Returns the Fingerprint of the molecule.

    Parameters
    ----------
    molecule : rdkit.Chem.rdchem.Mol
        A molecule.

    fpformat : str
        A valid fingerprint format.

    Returns
    -------
    Fingerprint
        rdkit.DataStructs.cDataStructs.ExplicitBitVect
    """
    assert isinstance(molecule, Chem.rdchem.Mol)
    assert isinstance(fpformat, str)
    fp = None

    if fpformat not in fps:
        raise AssertionError("'%s' is not a valid fingerprint format" % fpformat)
    if fpformat == "maccs":
        fp = MACCSkeys.FingerprintMol(molecule)
    elif fpformat == "morgan2":
        fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 2)
    elif fpformat == "morgan3":
        fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 3)
    elif fpformat == "morgan4":
        fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 4)
    elif fpformat == "morgan5":
        fp = AllChem.GetMorganFingerprintAsBitVect(molecule, 5)

    return fp


def fingerprint_to_bits(fp, bits=1024):
    """
    Converts a pybel.Fingerprint into a binary array

    Parameters
    ----------
    fp : rdkit.DataStructs.cDataStructs.ExplicitBitVect
        A fingerprint molecule.
    bits : int
        Number of bits (default is 1024)
    Returns
    -------
    bitarray
    """
    bits_list = bitarray(bits)
    bits_list.setall(0)

    for i in range(fp.GenNumBits()):
        if fp.GetBit(i):
            bits_list[i - 1] = 1

    return bits_list


def maximum_common_substructure(reference, molecule, match_rings=True, match_fraction=0.6, timeout=None):
    """
    Returns the Maximum Common Substructure (MCS) between two molecules.

    Parameters
    ----------
    reference : rdkit.Chem.Mol
        A molecule.
    molecule : rdkit.Chem.Mol
        Another molecule.
    match_rings : bool
        Force ring structure to match
    match_fraction : float
        Match is fraction of the reference atoms (default: 0.6)
    timeout: int
        Time out in seconds.
    Returns
    -------
    rdkit.Chem.MCS.MCSResult
        Maximum Common Substructure result.
    """

    assert isinstance(reference, rdkit.Chem.rdchem.Mol)

    min_num_atoms = math.ceil(reference.GetNumAtoms()) * match_fraction

    return MCS.FindMCS([reference, molecule], ringMatchesRingOnly=match_rings,
                       minNumAtoms=min_num_atoms, timeout=timeout,
                       atomCompare='any', bondCompare="any")


def mcs_similarity(mcs_result, molecule, atoms_weight=0.5, bonds_weight=0.5):
    """
    Returns the Maximum Common Substructure (MCS) between two molecules.

    $$ atoms\_weight * (mcs_res.similar\_atoms/mol.num_atoms) + bonds\_weight *
    (mcs_res.similar\_bonds/mol.num\_bonds) $$

    Parameters
    ----------
    mcs_result : rdkit.Chem.MCS.MCSResult
        The result of a Maximum Common Substructure run.
    molecule : rdkit.Chem.Mol
        A molecule.
    atoms_weight : float
        How much similar atoms matter.
    bonds_weight : float
        How much similar bonds matter.
    Returns
    -------
    float
        Similarity value
    """

    assert isinstance(mcs_result, MCS.MCSResult)
    assert isinstance(molecule, Chem.rdchem.Mol)

    atoms_score = atoms_weight * (float(mcs_result.numAtoms) / float(molecule.GetNumAtoms()))
    try:
        bonds_score = bonds_weight * (float(mcs_result.numBonds) / float(molecule.GetNumBonds()))
    except ZeroDivisionError:
        bonds_score = .0
    return atoms_score + bonds_score


def structural_similarity(reference, molecule, atoms_weight=0.5, bonds_weight=0.5,
                          match_rings=True, match_fraction=0.6, timeout=None):
    """
    Returns a structural similarity based on the Maximum Common Substructure (MCS) between two molecules.

    $$ mcs\_s(ref) * mcs\_s(mol) $$

    Parameters
    ----------
    reference : rdkit.Chem.Mol
        The result of a Maximum Common Substructure run.
    molecule : rdkit.Chem.Mol
        A molecule.
    atoms_weight : float
        How much similar atoms matter.
    bonds_weight : float
        How much similar bonds matter.
    match_rings : bool
        Force ring structure to match.
    match_fraction : float
        Match is fraction of the reference atoms (default: 0.6).
    timeout : int
        Time out in seconds.

    Returns
    -------
    float
        Similarity between reference and molecule.
    """

    reference = Chem.RemoveHs(reference, implicitOnly=True, updateExplicitCount=True)
    molecule = Chem.RemoveHs(molecule, implicitOnly=True, updateExplicitCount=True)

    mcs_res = maximum_common_substructure(reference, molecule, match_rings=match_rings,
                                          match_fraction=match_fraction, timeout=timeout)

    ref_similarity = mcs_similarity(mcs_res, reference, atoms_weight=atoms_weight, bonds_weight=bonds_weight)
    mol_similarity = mcs_similarity(mcs_res, molecule, atoms_weight=atoms_weight, bonds_weight=bonds_weight)
    return ref_similarity * mol_similarity


def monte_carlo_volume(molecule, coordinates=None, tolerance=1, max_iterations=10000, step_size=1000,
                       seed=time.time(), verbose=False, forcefield='mmff94', steps=100):
    """
    Adapted from:

    Simple Monte Carlo estimation of VdW molecular volume (in A^3)
    by Geoffrey Hutchison <geoffh@pitt.edu>

    https://github.com/ghutchis/hutchison-cluster

    Parameters
    ----------
    molecule : rdkit.Chem.rdchem.Mol
        A molecule from rdkit.
    coordinates : list
        A list of pre selected x,y,z coords. It must match the atoms order.
    tolerance : float
        The tolerance for convergence of the monte carlo abs(new_volume - volume) < tolerance
    max_iterations : int
        Number of iterations before the algorithm starts.
    step_size : int
        Number of points to add each step.
    seed : object
        A valid seed for random.
    verbose : bool
        Print debug information if True.
    forcefield : str
        The force field to get a 3D molecule. (only if it is not 3D already)
    steps : int
        The number of steps used for the force field to get a 3D molecule. (only if it is not 3D already)

    Returns
    -------
    float
        Molecule volume
    """

    assert isinstance(molecule, rdkit.Chem.rdchem.Mol)
    if coordinates is None:
        if len(molecule.GetConformers()) == 0:
            AllChem.EmbedMolecule(molecule)
            AllChem.UFFOptimizeMolecule(molecule)

        conformer = molecule.GetConformer(0)
        coordinates = []

        for index, atom in enumerate(molecule.GetAtoms()):
            pos = tuple(conformer.GetAtomPosition(index))
            coordinates.append(pos)

    vdw_radii = []
    for index, atom in enumerate(molecule.GetAtoms()):
        radius = periodic_table.GetRvdw(atom.GetAtomicNum())
        vdw_radii.append(radius)

    coordinates = np.array(coordinates, dtype=np.float32)
    vdw_radii = np.array(vdw_radii, dtype=np.float32)
    return mc_vol(coordinates, vdw_radii, tolerance, max_iterations, step_size, seed, int(verbose))



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from cameo.flux_analysis.analysis import find_essential_metabolites
from cameo.flux_analysis.simulation import pfba
from cameo.strain_design.heuristic.evolutionary.decoders import SetDecoder
from cameo.strain_design.heuristic.evolutionary.evaluators import TargetEvaluator
from cameo.strain_design.heuristic.evolutionary.optimization import TargetOptimization
from cobra.exceptions import OptimizationError

from marsi.cobra.flux_analysis.manipulation import knockout_metabolite
from marsi.utils import search_metabolites

logger = logging.getLogger(__name__)

__all__ = ["MetaboliteKnockoutOptimization"]

METABOLITE_KNOCKOUT_TYPE = "metabolite knockout"


class MetaboliteDecoder(SetDecoder):
    """
    Decoder for set representation. Converts an integer set into metabolites.

    Parameters
    ----------

    representation : list
        Reactions.
    model : SolverBasedModel

    """

    def __init__(self, representation, model, *args, **kwargs):
        super(MetaboliteDecoder, self).__init__(representation, model, *args, **kwargs)

    def __call__(self, individual, flat=False, **kwargs):
        """
        Parameters
        ----------

        **kwargs
        individual : list
            a list of integers
        flat : bool
            if True, returns strings. Otherwise returns Reaction.

        Returns
        -------
        list
            Decoded representation

        """
        metabolites = tuple(self.representation[index] for index in individual)
        return metabolites


class AntiMetaboliteEvaluator(TargetEvaluator):
    def __init__(self, essential_metabolites=None, inhibition_fraction=.0, competition_fraction=.0, *args, **kwargs):
        super(AntiMetaboliteEvaluator, self).__init__(*args, **kwargs)
        self.essential_metabolites = essential_metabolites or []
        self.inhibition_fraction = inhibition_fraction
        self.competition_fraction = competition_fraction

    def _evaluate_individual(self, individual):
        return self.evaluate_individual(individual)

    def evaluate_individual(self, individual):
        from marsi.cobra.flux_analysis.manipulation import apply_anti_metabolite

        specie_ids = self.decoder(individual)
        metabolite_targets = [search_metabolites(self.model, val) for val in specie_ids]
        with self.model:
            for metabolites in metabolite_targets:
                apply_anti_metabolite(self.model, metabolites, self.essential_metabolites,
                                      self.simulation_kwargs['reference'], self.inhibition_fraction,
                                      self.competition_fraction)

            try:
                solution = self.simulation_method(self.model, **self.simulation_kwargs)
                return self.objective_function(self.model, solution, metabolite_targets)
            except OptimizationError:
                return self.objective_function.worst_fitness()


class MetaboliteKnockoutEvaluator(TargetEvaluator):
    def __init__(self, essential_metabolites=None, *args, **kwargs):
        super(MetaboliteKnockoutEvaluator, self).__init__(*args, **kwargs)
        self.essential_metabolites = essential_metabolites or []

    def _evaluate_individual(self, individual):
        return self.evaluate_individual(individual)

    def evaluate_individual(self, individual):

        specie_ids = self.decoder(individual)
        metabolite_targets = [search_metabolites(self.model, val) for val in specie_ids]
        with self.model:
            for metabolites in metabolite_targets:
                for metabolite in metabolites:
                    knockout_metabolite(self.model, metabolite, ignore_transport=True, allow_accumulation=True)
            try:
                solution = self.simulation_method(self.model, **self.simulation_kwargs)
                return self.objective_function(self.model, solution, metabolite_targets)
            except OptimizationError:
                return self.objective_function.worst_fitness()


class MetaboliteKnockoutOptimization(TargetOptimization):
    """
    Knockout optimization using metabolites.

    Attributes
    ----------
    model : SolverBasedModel
        A constraint-based model.
    heuristic_method : inspyred.ec.EvolutionaryComputation
        An evolutionary algorithm.
    objective_function : objective function or list(objective function)
        The objectives for the algorithm to maximize.
    seed : int
        A seed for random. It is auto-generated if None is given.
    termination : inspyred.ec.terminators
        A termination criteria for the algorithm. The default is inspyred.ec.terminators.evaluation_termination.
    simulation_method: flux_analysis.simulation
        The method used to simulate the model.
    wt_reference: dict
        A reference initial state for the optimization. It is required for flux_analysis.simulation.lmoma and
        flux_analysis.simulation.room. If not given, it will be computed using flux_analysis.simulation.pfba
    metabolites: list
        A list of valid metabolites to knockout. If None, then all metabolites in the model will be knockout candidates
        except the ones defined in essential_metabolites
    essential_metabolites: list
        A list of metabolites that cannot be knocked out. If None, then all essential genes will be removed from the
        valid genes set.

    Methods
    -------
    run(view=config.default_view, maximize=True, **kwargs)


    See Also
    --------
    inspyred
    cameo.config.default_view

    Example
    -------
    >>> from cameo import models
    >>> model = models.bigg.iJO1366
    >>> from cameo.strain_design.heuristic.evolutionary.objective_functions import biomass_product_coupled_yield
    >>> bpcy = biomass_product_coupled_yield(model.reactions.Ec_biomass_iJO1366_core_53p95,
    >>>                                      model.reactions.EX_succ_e,
    >>>                                      model.reactions.EX_glc__D_e)
    >>> knockout_optimization = MetaboliteKnockoutOptimization(model=model, objective_function=bpcy)
    >>> knockout_optimization.run(max_evaluations=50000)

    """
    def __init__(self, metabolites=None, essential_metabolites=None, n_carbons=2, compartments="c",
                 skip_essential_metabolites=False, *args, **kwargs):
        super(MetaboliteKnockoutOptimization, self).__init__(*args, **kwargs)

        if compartments is None:
            compartments = list(self.model.compartments.keys())
        self.compartments = compartments

        if metabolites is None:
            self.metabolites = set([met.id[:-2] for met in self.model.metabolites if
                                    met.elements.get('C', 0) >= n_carbons and met.compartment in self.compartments])
        else:
            self.metabolites = metabolites

        logger.debug("Computing essential reactions...")
        if skip_essential_metabolites:
            self.essential_metabolites = set()
        else:
            _essential_metabolites = find_essential_metabolites(self.model)
            self.essential_metabolites = set([m.id for m in _essential_metabolites])

        if essential_metabolites:
            self.essential_metabolites.update(essential_metabolites)

        self.representation = list(self.metabolites)
        self._target_type = METABOLITE_KNOCKOUT_TYPE
        self._decoder = MetaboliteDecoder(self.representation, self.model)
        self._evaluator = MetaboliteKnockoutEvaluator(
            model=self.model,
            decoder=self._decoder,
            objective_function=self.objective_function,
            simulation_method=self._simulation_method,
            simulation_kwargs=self._simulation_kwargs,
            essential_metabolites=self.essential_metabolites)

    @TargetOptimization.simulation_method.setter
    def simulation_method(self, simulation_method):
        if self._simulation_kwargs.get("reference", None) is None:
            logger.warning("No WT reference found, generating using pfba.")
            self._simulation_kwargs['reference'] = pfba(self.model).fluxes
            logger.warning("Reference successfully computed.")
        self._simulation_method = simulation_method

    @TargetOptimization.simulation_kwargs.setter
    def simulation_kwargs(self, simulation_kwargs):
        if self._simulation_method and simulation_kwargs.get("reference", None) is None:
            logger.warning("No WT reference found, generating using pfba.")
            simulation_kwargs['reference'] = pfba(self.model).fluxes
            logger.warning("Reference successfully computed.")
        self._simulation_kwargs = simulation_kwargs



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
BiGG database API v2
"""
import os
import requests

BASE_URL = "http://bigg.ucsd.edu/api/v2/"


class DBVersion(object):
    """
    Version container.

    Attributes
    ----------
    version : str
        Database version.
    api_version : str
        Version of the API used.
    last_update : str
        Last update timestamp.

    """

    def __init__(self, version, api_version, last_update):
        self.version = version
        self.api_version = api_version
        self.last_update = last_update

    def __repr__(self):
        return "BiGG Database version %s (API %s). Last update: %s" % (self.version, self.api_version, self.last_update)


def database_version():
    """
    Retrieves the current version of BiGG database
    """
    response = requests.get(BASE_URL + "database_version")
    response.raise_for_status()

    data = response.json()
    return DBVersion(data['bigg_models_version'], data['api_version'], data['last_updated'])


def download_model(model_id, file_format="json", save=True, path="."):
    """
    Download models from BiGG. You can chose to save the file or to return the JSON data.

    Parameters
    ----------

    model_id: str
        A valid id for a model in BiGG.
    file_format: str
        If you want to save the file, you can import the model in the following formats:
            1. json (JSON format)
            2. xml (SBML)
            3. xml.gz (SBML compressed)
            4. mat (MATLAB)
    save: bool
        If True, writes the model to a file with the model name (the path can be specified).
    path: str
        Specifies in which folder the model should be written if *save* is True.
    """

    if save:
        response = requests.get("http://bigg.ucsd.edu/static/models/%s.%s" % (model_id, file_format), stream=True)
        response.raise_for_status()
        with open(os.path.join(path, "%s.%s" % (model_id, file_format)), "wb") as model_file:
            for block in response.iter_content(1024):
                model_file.write(block)
    else:
        response = requests.get(BASE_URL + "models/%s" % model_id)
        response.raise_for_status()
        return response.json()


def model_details(model_id):
    """
    Summarize the model.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s" % model_id)
    response.raise_for_status()
    return response.json()


def list_models():
    """
    Lists all models available in BiGG.
    """
    response = requests.get(BASE_URL + "models/")
    response.raise_for_status()
    return response.json()


def list_reactions():
    """
    List all reactions available in BiGG.
    """
    response = requests.get(BASE_URL + "universal/reactions")
    response.raise_for_status()
    return response.json()


def list_model_reactions(model_id):
    """
    List all reactions in a model.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/reactions" % model_id)
    response.raise_for_status()
    return response.json()


def get_reaction(reaction_id):
    """
    Retrieve a reaction from BiGG.

    Parameters
    ----------
    reaction_id: str
        A valid id for a reaction in BiGG.
    """
    response = requests.get(BASE_URL + "universal/reactions/%s" % reaction_id)
    response.raise_for_status()
    return response.json()


def get_model_reaction(model_id, reaction_id):
    """
    Retrieve a reaction in the context of a model from BiGG.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    reaction_id: str
        A valid id for a reaction in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/reactions/%s" % (model_id, reaction_id))
    response.raise_for_status()
    return response.json()


def list_metabolites():
    """
    List all metabolites in BiGG.
    """
    response = requests.get(BASE_URL + "universal/metabolites")
    response.raise_for_status()
    return response.json()


def list_model_metabolites(model_id):
    """
    List all metabolites in a model.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/metabolites" % model_id)
    response.raise_for_status()
    return response.json()


def get_metabolite(metabolite_id):
    """
    Retrieve a metabolite from BiGG.

    Parameters
    ----------
    metabolite_id: str
        A valid id for a reaction in BiGG.
    """
    response = requests.get(BASE_URL + "universal/metabolites/%s" % metabolite_id)
    response.raise_for_status()
    return response.json()


def get_model_metabolite(model_id, metabolite_id):
    """
    Retrieve a metabolite in the context of a model from BiGG.

    Parameters
    ----------
    metabolite_id: str
        A valid id for a reaction in BiGG.
    model_id: str
        A valid id for a model in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/metabolites/%s" % (model_id, metabolite_id))
    response.raise_for_status()
    return response.json()


def list_model_genes(model_id):
    """
    List all genes in a model.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/genes" % model_id)
    response.raise_for_status()
    return response.json()


def get_model_gene(model_id, gene_id):
    """
    Retrieve a gene in the context of a model from BiGG.

    Parameters
    ----------
    model_id: str
        A valid id for a model in BiGG.
    gene_id: str
        A valid id for a gene in BiGG.
    """
    response = requests.get(BASE_URL + "models/%s/metabolites/%s" % (model_id, gene_id))
    if response.ok:
        return response.json()


def search(query, search_type):
    """
    Fuzzy search the BiGG database.

    Parameters
    ----------
    query: str
        Whatever you are searching for (ids, names, etc...).
    search_type: str
        Search domain. One of "models", "genes", "reactions", "metabolites".

    """
    response = requests.get(BASE_URL + "search", params=dict(query=query, search_type=search_type))
    response.raise_for_status()
    return response.json()



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cement.core.foundation import CementApp

from marsi.cli.controllers import MarsiBaseController
from marsi.cli.controllers.chemistry import ChemistryController
from marsi.cli.controllers.database import DatabaseController
from marsi.cli.controllers.modeling import OptimizationController


class MarsiApp(CementApp):
    class Meta:
        label = 'marsi'
        base_controller = 'base'
        handlers = [
            MarsiBaseController,
            DatabaseController,
            OptimizationController,
            ChemistryController
        ]



# Copyright 2016 Chr. Hansen A/S
# Copyright 2016 The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

from cameo.core.utils import get_reaction_for
from cameo.flux_analysis.analysis import find_essential_metabolites
from cameo.flux_analysis.simulation import fba
from cameo.strain_design import OptKnock, OptGene, DifferentialFVA
from cameo.strain_design.heuristic.evolutionary.objective_functions import biomass_product_coupled_yield
from cobra.core.model import Model
from pandas import DataFrame

from marsi.cobra.strain_design.evolutionary import OptMet
from marsi.cobra.strain_design.post_processing import replace_design
from marsi.cobra.utils import CURRENCY_METABOLITES


class GenericMARSIDesignMethod(object):
    """
    Generic wrapper for Metabolite Analog design method.

    This one just runs a optimization method

    Example
    -------
    >>> from marsi.cobra import strain_design
    >>> from cameo import models
    >>> from cameo.strain_design import OptGene
    >>> designer = strain_design.GenericMARSIDesignMethod(model=models.bigg.iJO1366)
    >>> designer.optimize_with_reaction("succ_e", max_interventions=5, substrate="EX_glc__D_e",
    >>> biomass="BIOMASS_Ec_iJO1366_core_53p95M", max_results=25, design_method=OptGene, manipulation_type="reactions")


    """
    def __init__(self, model=None, nearest_neighbors_model=None, min_tanimoto=0.75, currency_metabolites=None,
                 essential_metabolites=None):
        assert isinstance(model, Model)
        self.model = model
        self.nearest_neighbors_model = nearest_neighbors_model
        self.min_tanimoto = min_tanimoto
        self.currency_metabolites = currency_metabolites or CURRENCY_METABOLITES
        if essential_metabolites is None:
            essential_metabolites = []

        self.essential_metabolites = find_essential_metabolites(model) + essential_metabolites
        for i, m in enumerate(self.essential_metabolites):
            if isinstance(m, str):
                self.essential_metabolites[i] = model.metabolites.get_by_id(m)

    def optimize_with_reaction(self, target, max_interventions=1, substrate=None,
                               biomass=None, design_method=OptKnock, max_results=100,
                               non_essential_metabolites=False, max_evaluations=20000, **design_kwargs):
        raise NotImplementedError

    def optimize_with_metabolites(self, target, max_interventions=1, substrate=None, biomass=None,
                                  max_results=100, non_essential_metabolites=False, max_evaluations=20000,
                                  **design_kwargs):
        raise NotImplementedError

    def essential_metabolites_reactions(self):
        essential_metabolites = find_essential_metabolites(self.model)
        reactions = set()
        for metabolite in essential_metabolites:
            reactions.update(metabolite.reactions)

        return reactions

    def _evaluate_designs(self, strain_designs, objective_function):
        evaluated_designs = DataFrame(columns=["design", "fitness"])

        for i, design in enumerate(strain_designs):
            with self.model:
                design.apply(self.model)
                solution = fba(self.model)
                fitness = objective_function(self.model, solution, design.targets)
                evaluated_designs.loc[i] = [design, fitness]

        return evaluated_designs


class RandomMutagenesisDesign(GenericMARSIDesignMethod):
    """
    Apply only knockout like designs where total loss of functions are expected.



    """
    def optimize_with_reaction(self, target, max_interventions=1, substrate=None,
                               biomass=None, design_method="optgene", max_results=100,
                               non_essential_metabolites=False, max_evaluations=20000, **designer_kwargs):

        target_flux = get_reaction_for(self.model, target)

        exclude_reactions = []
        if non_essential_metabolites:
            exclude_reactions = self.essential_metabolites_reactions()

        if 'essential_reactions' in designer_kwargs:
            designer_kwargs['essential_reactions'] += exclude_reactions
        else:
            designer_kwargs['essential_reactions'] = exclude_reactions

        bpcy = biomass_product_coupled_yield(biomass, target_flux, substrate)

        if design_method == "optgene":
            designer = OptGene(self.model, **designer_kwargs)
            knockouts = designer.run(max_knockouts=max_interventions, biomass=biomass, substrate=substrate,
                                     target=target_flux, max_results=max_results, max_evaluations=max_evaluations,
                                     use_nullspace_simplification=False)
        elif design_method == "optknock":
            designer = OptKnock(self.model, **designer_kwargs)
            knockouts = designer.run(max_knockouts=max_interventions, biomass=biomass, substrate=substrate,
                                     target=target_flux, max_results=max_results)
        else:
            raise ValueError("'design_method' can be one of 'optgene' or 'optknock'")

        designs = self._evaluate_designs(iter(knockouts), bpcy)

        anti_metabolites_design = DataFrame()
        for i_, row in designs.iterrows():
            _anti_metabolites_design = replace_design(self.model, row.design, row.fitness, bpcy, fba, {},
                                                      ignore_metabolites=self.currency_metabolites,
                                                      essential_metabolites=self.essential_metabolites)
            anti_metabolites_design = anti_metabolites_design.append(_anti_metabolites_design, ignore_index=True)

        return anti_metabolites_design

    def optimize_with_metabolites(self, target, max_interventions=1, substrate=None, biomass=None,
                                  max_results=100, max_evaluations=20000, **design_kwargs):

        target_flux = get_reaction_for(self.model, target)

        designer = OptMet(model=self.model, essential_metabolites=CURRENCY_METABOLITES, **design_kwargs)
        knockouts = designer.run(max_knockouts=max_interventions, biomass=biomass, substrate=substrate,
                                 target=target_flux, max_results=max_results, max_evaluations=max_evaluations)

        return knockouts.data_frame


class ALEDesign(GenericMARSIDesignMethod):
    """
    Apply both knockout and flux modulation. The strains will be selected via growth rate proxy.


    """

    def optimize_with_reaction(self, target, max_interventions=1, substrate=None,
                               biomass=None, design_method="differential_fva", max_results=100,
                               non_essential_metabolites=False, **designer_kwargs):

        target_flux = get_reaction_for(self.model, target)
        bpcy = biomass_product_coupled_yield(biomass, target_flux, substrate)

        designer = DifferentialFVA(self.model, objective=target, variables=[biomass],
                                   points=max_results, **designer_kwargs)

        evaluated_designs = DataFrame(columns=["strain_designs", "fitness"])
        for i, design in enumerate(designer.run()):
            with self.model:
                design.apply(self.model)
                solution = self.model.optimize()
                fitness = bpcy(self.model, solution, design.targets)
                evaluated_designs.loc[i] = [design, fitness]

        anti_metabolites_design = DataFrame()
        for _, row in evaluated_designs.iterrows():
            _anti_metabolites_design = replace_design(self.model, row.design, row.fitness, bpcy, fba, {},
                                                      ignore_metabolites=self.currency_metabolites,
                                                      essential_metabolites=self.essential_metabolites)
            anti_metabolites_design = anti_metabolites_design.append(_anti_metabolites_design, ignore_index=True)

        return anti_metabolites_design

    def optimize_with_metabolites(self, target, max_interventions=1, substrate=None, biomass=None,
                                  max_results=100, max_evaluations=20000, **design_kwargs):
        raise NotImplementedError



"""add solubility

Revision ID: ef39a4ae2c8c
Revises: c1b678211e9d
Create Date: 2017-04-21 12:30:00.211355

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'ef39a4ae2c8c'
down_revision = 'c1b678211e9d'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('metabolites', sa.Column('solubility', sa.Float))


def downgrade():
    op.drop_column('metabolites', 'solubility')



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from pandas import DataFrame

from marsi.utils import internal_data_dir

try:
    bigg_metabolites = DataFrame.from_csv(os.path.join(internal_data_dir, "bigg_models_metabolites.txt"), sep="\t")
    bigg_metabolites.database_links = bigg_metabolites.database_links.apply(eval)
    bigg_metabolites.model_list = bigg_metabolites.model_list.apply(str.split, args=(", ",))
except IOError:
    bigg_metabolites = DataFrame()



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import getpass
import logging
import os

import six
from openbabel import obErrorLog, obError, obWarning, obInfo, obDebug
from sqlalchemy.event import listens_for
from sqlalchemy.exc import DisconnectionError

__all__ = ['Level', 'log', 'prj_dir', 'db_url']

TRAVIS = os.environ.get("TRAVIS", False)


logger = logging.getLogger(__name__)


def _add_process_guards(engine):
    """Add multiprocessing guards.

    Forces a connection to be reconnected if it is detected
    as having been shared to a sub-process.

    """

    @listens_for(engine, "connect")
    def connect(connection, connection_record):
        connection_record.info['pid'] = os.getpid()

    @listens_for(engine, "checkout")
    def checkout(connection, connection_record, connection_proxy):
        pid = os.getpid()
        if connection_record.info['pid'] != pid:
            connection_record.connection = connection_proxy.connection = None
            raise DisconnectionError(
                "Connection record belongs to pid %s, "
                "attempting to check out in pid %s" %
                (connection_record.info['pid'], pid))


class Level:
    ERROR = obError
    WARNING = obWarning
    INFO = obInfo
    DEBUG = obDebug


class LogConf(object):
    __level_to_logger = {
        Level.ERROR: logging.ERROR,
        Level.WARNING: logging.WARNING,
        Level.DEBUG: logging.DEBUG,
        Level.INFO: logging.INFO
    }

    def __init__(self):
        self._app_logger = logging.getLogger('marsi.*')
        self._level = Level.WARNING

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level):
        obErrorLog.SetOutputLevel(level)
        self._app_logger.setLevel(self.__level_to_logger[level])
        self._level = level


log = LogConf()

log.level = Level.INFO

default = {'marsi': {
    'prj_dir': "%s/.marsi" % os.getenv('HOME'),
    'db_name': 'marsi-db',
    'db_engine': "postgresql",
    'db_user': getpass.getuser(),
    'db_host': "localhost",
    'db_port': 5432,
    'db_pass': None}
}


def get_default(config, section, key, default):
    try:
        return config.get(section, key)
    except Exception:
        return default[section].get(key, None)


config = six.moves.configparser.ConfigParser()

# TODO: specify database connection configuration

try:
    logger.debug("Looking for setup.cfg")
    with open('setup.cfg') as file:
        config.readfp(file)
except IOError as e:
    logger.debug("Not available %s" % str(e))
    config = default

try:
    if isinstance(config, dict) or six.PY3:
        prj_dir = os.path.abspath(config['marsi'].get('prj_dir'))
        db_name = config['marsi'].get(['db_name'])

    else:
        prj_dir = os.path.abspath(get_default(config, 'marsi', 'prj_dir', default))
        db_name = get_default(config, 'marsi', 'db_name', default)
except Exception:
    prj_dir = default['marsi']['prj_dir']
    db_name = default['marsi']['db_name']

logger.info("Working dir %s" % prj_dir)

if not os.path.exists(prj_dir):
    os.mkdir(prj_dir)


log.level = Level.ERROR

try:
    if isinstance(config, dict) or six.PY3:
        db_engine = config['marsi'].get('db_engine', "postgresql")
    else:
        db_engine = get_default(config, 'marsi', 'db_engine', default)

    if TRAVIS:
        username = 'postgres'
        password = None
        host = 'localhost'
        port = 5432
    else:  # TODO: needs documentation
        if isinstance(config, dict) or six.PY3:
            username = config['marsi'].get('db_user', getpass.getuser())
            password = config['marsi'].get('db_pass', None)
            host = config['marsi'].get("db_host", "localhost")
            try:
                port = int(config['marsi'].get("db_port", 5432))
            except ValueError:
                port = None
        else:
            username = get_default(config, 'marsi', 'db_user', default)
            password = get_default(config, 'marsi', 'db_pass', default)
            host = get_default(config, 'marsi', "db_host", default)
            try:
                port = int(get_default(config, 'marsi', "db_port", default))
            except ValueError:
                port = None

    if db_engine == 'sqlite':
        db_url = "sqlite:///%s" % db_name
    else:
        if password is None:
            user_access = username
        else:
            user_access = "%s:%s" % (username, password)

        if host is None:
            host_port = None
        elif port is None:
            host_port = host
        else:
            host_port = "%s:%i" % (host, port)

        if host_port is None:
            db_url = "%s://%s/%s" % (db_engine, user_access, db_name)
        else:
            db_url = "%s://%s@%s/%s" % (db_engine, user_access, host_port, db_name)

except Exception as e:
    print(e)
    default_session = None
    engine = None
    db_url = None
    logger.warning("You are running MARSI without a database connection. \n"
                   "You will not be able to use many functionalities including: \n"
                   "1. Query the metabolite database\n"
                   "2. Search for metabolite analogs")
else:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker

    try:
        engine = create_engine(db_url, client_encoding='utf8', pool_size=10)
    except TypeError:
        # The pool_size argument won't work for the default SQLite setup in SQLAlchemy 0.7, try without
        engine = create_engine(db_url)

    _add_process_guards(engine)

    Session = sessionmaker(engine)
    default_session = Session()



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from sklearn import neighbors

from cameo.parallel import SequentialView
from sqlalchemy.orm import load_only

from marsi.io.db import Metabolite

try:  # pragma: no cover
    import pyopencl as cl
    MF = cl.mem_flags
    cl_available = True
except ImportError:
    class ClFail:
        def __getattr__(self, item):
            raise RuntimeError("OpenCL is not available. try 'pip install pyopencl'")

    cl = ClFail()
    MF = None
    cl_available = False

import numpy as np
from pandas import DataFrame

from marsi.utils import timing
from marsi.nearest_neighbors import model_ext

logger = logging.getLogger(__name__)

# Tanimoto coefficient calculation is implemented based on to OpenBabel's implementation
# https://github.com/openbabel/openbabel/blob/master/include/openbabel/fingerprint.h#L86
nn_source = """
__kernel void nn(__global float *distances, __global int *fingerprint, __constant int *fpLength,
                 __global int *database, __global int *positions, __constant int *lengths) {
    int gid = (int)get_global_id(0);

    float tanimotoCoefficient = -1;

    int dbEntryLength = lengths[gid];
    int fpLen = fpLength[0];

    if (fpLen == dbEntryLength) {
        int andBits = 0;
        int orBits = 0;
        int fpPos = positions[gid];
        int i;
        for (i=0; i < fpLen; ++i) {
            int fpAnd = database[fpPos+i] & fingerprint[i];
            int fpOr = database[fpPos+i] | fingerprint[i];
            andBits += popcount(fpAnd);
            orBits += popcount(fpOr);
        }

        tanimotoCoefficient = (float)andBits/(float)orBits;
    }
    distances[gid] = 1 - tanimotoCoefficient;
}
"""


class KNN(object):
    """
    K-Nearest Neighbors runner object.

    It is assigned to a model and runs the `knn` function.

    Attributes
    ----------
    fp : numpy.array
        A numpy.array with the fingerprint values.
    k : int
        The maximum number of neighbors to retrieve.
    mode : str
        'native' to run python implementation or 'cl' to run OpenCL implementation if available.
    """
    def __init__(self, fingerprint, k, mode):
        self.fp = np.array(list(fingerprint), dtype=np.int32)
        self.k = k
        self.mode = mode

    def __call__(self, nn):
        return nn.knn(self.fp, k=self.k)

    def __getstate__(self):
        return dict(fp=self.fp.tolist(), k=self.k, mode=self.mode)

    def __setstate__(self, d):
        d['fp'] = np.array(d['fp'], dtype=np.int32)
        self.__dict__.update(d)


class RNN(object):
    """
    R-Nearest Neighbors runner object.

    It is assigned to a model and runs the `rnn` function.

    Attributes
    ----------
    fp : numpy.array
        A numpy.array with the fingerprint values.
    radius : float
        A distance radius ]0, 1].
    mode : str
        'native' to run python implementation or 'cl' to run OpenCL implementation if available.
    """
    def __init__(self, fingerprint, radius, mode):
        self.fp = np.array(list(fingerprint), dtype=np.int32)
        assert 0 < radius <= 1
        self.radius = radius
        self.mode = mode

    def __call__(self, nn):
        return nn.rnn(self.fp, radius=self.radius, mode=self.mode)

    def __getstate__(self):
        return dict(fp=self.fp.tolist(), radius=self.radius, mode=self.mode)

    def __setstate__(self, d):
        d['fp'] = np.array(d['fp'], dtype=np.int32)
        self.__dict__.update(d)


class Distance(object):
    """
    Distance runner object.

    It is assigned to a model and runs the `distance` function.

    Attributes
    ----------
    fp : numpy.array
        A numpy.array with the fingerprint values.
    mode : str
        'native' to run python implementation or 'cl' to run OpenCL implementation if available.
    """
    def __init__(self, fingerprint, mode):
        self.fp = np.array(list(fingerprint), dtype=np.int32)
        self.mode = mode

    def __call__(self, nn):
        return nn.distances(self.fp, mode=self.mode)

    def __getstate__(self):
        return dict(fp=self.fp, mode=self.mode)

    def __setstate__(self, d):
        d['fp'] = np.array(d['fp'], dtype=np.int32)
        self.__dict__.update(d)


class DistributedNearestNeighbors(object):
    """
    Nearest Neighbors distributed implementation.

    Attributes
    ----------
    index : numpy.array
        The index of all entries across multiple models.
    """
    def __init__(self, nns):
        self._nns = nns

    def k_nearest_neighbors(self, fingerprint, k=5, mode="native", view=SequentialView()):
        """
        Retrieves the K nearest neighbors to a fingerprint.

        Parameters
        ----------
        fingerprint : list, np.array, tuple
            A fingerprint to use as query.
        k : int
            The number of neighbors to retrieve.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.
        view : cameo.parallel.ParallelView, cameo.parallel.SequentialView
            A parallel mode runner.
        Returns
        -------
        dict
            A dictionary with the InChI Key as key and the distance as value.
        """
        func = KNN(fingerprint, k, mode)
        results = view.map(func, self._nns)
        neighbors = {}
        [neighbors.update(res) for res in results]
        return dict(sorted(neighbors.items(), key=lambda x: x[1])[:k])

    def radius_nearest_neighbors(self, fingerprint, radius=0.25, mode="native", view=SequentialView()):
        """
        Retrieves the nearest neighbors to a fingerprint within a distance radius.

        Parameters
        ----------
        fingerprint : list, np.array, tuple
            A fingerprint to use as query.
        radius : float
            A distance radius ]0, 1].
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.
        view : cameo.parallel.ParallelView, cameo.parallel.SequentialView
            A parallel mode runner.
        Returns
        -------
        dict
            A dictionary with the InChI Key as key and the distance as value.
        """
        func = RNN(fingerprint, radius, mode)
        results = view.map(func, self._nns)
        neighbors = {}
        [neighbors.update(res) for res in results]
        return neighbors

    def distances(self, fingerprint, mode="native", view=SequentialView()):
        """
        Retrieves the distance a fingerprint and all elements in the model.

        Parameters
        ----------
        fingerprint : list, np.array, tuple
            A fingerprint to use as query.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.
        view : cameo.parallel.ParallelView, cameo.parallel.SequentialView
            A parallel mode runner.
        Returns
        -------
        dict
            A dictionary with the InChI Key as key and the distance as value.
        """
        func = Distance(fingerprint, mode)
        results = view.map(func, self._nns)
        distances = {}
        [distances.update(res) for res in results]
        return distances

    def __repr__(self):
        return " | ".join(repr(nn) for nn in self._nns)

    def __getitem__(self, index):
        return self._nns[index]

    @property
    def index(self):
        return np.concatenate([nn.index for nn in self._nns])

    def distance_matrix(self, mode="native"):
        """
        Generates a distance matrix between all elements in the models.

        Parameters
        ----------
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.

        Returns
        -------
        numpy.array
            The distance matrix.
        """
        index = self.index
        size = len(index)
        matrix = np.zeros((size, size), dtype=np.float)
        for i in range(size):
            matrix[i] = self.distances(self.feature(i), mode=mode)

        return matrix

    def feature(self, index):
        """
        Retrieves the fingerprint at a given index. The index is global for the ensemble of models.

        Returns
        -------
        numpy.array
            The fingerprint.

        Raises
        ------
        IndexError
        """
        total = len(self)
        if index >= total:
            raise IndexError(index)

        group = 0
        off = 0
        for nn in self._nns:
            if index + 1 > len(nn):
                group += 1
                off += len(nn)

        return self._nns[group][index - off]

    def __len__(self):
        return sum(len(nn) for nn in self._nns)


class NearestNeighbors(model_ext.CNearestNeighbors):
    def __init__(self, index, features, features_lengths, use_cl=False, opencl_context=None):
        features = np.concatenate(features).astype(np.int32)
        features_lengths = np.array(features_lengths, dtype=np.int32)
        super(NearestNeighbors, self).__init__(features, features_lengths)
        self._index = index
        self._use_cl = use_cl
        if cl_available and self._use_cl:
            if opencl_context is None:
                self._ctx = cl.create_some_context()
            else:
                self._ctx = opencl_context
            self._program = cl.Program(self._ctx, nn_source).build()

    @property
    def cl_context(self):
        if not self._use_cl:
            raise RuntimeError("Configuration is set to don't run CL")
        return self._ctx

    @cl_context.setter
    def cl_context(self, ctx):
        if not self._use_cl:
            raise RuntimeError("Configuration is set to don't run CL")
        self._ctx = ctx
        self._program = cl.Program(self._ctx, nn_source).build()

    def queue(self):
        return cl.CommandQueue(self.cl_context)

    @property
    def program(self):
        if not self._use_cl:
            raise RuntimeError("Configuration is set to don't run CL")
        return self._program

    def __getstate__(self):
        state = super(NearestNeighbors, self).__getstate__()
        state.update({"_index": self._index, "_use_cl": self._use_cl})
        return state

    def __getitem__(self, index):
        start = self.start_positions[index]
        length = self.start_positions[index]
        end = start + length
        return self._features[start:end]

    def __setstate__(self, state):
        super(NearestNeighbors, self).__setstate__(state)
        self._index = state['_index']
        self._use_cl = state['_use_cl']
        if cl_available and self._use_cl:
            self.cl_context = cl.create_some_context()

    def knn(self, fingerprint, k, mode="native"):
        """
        K-Nearest Neighbors

        Parameters
        ----------

        fingerprint : ndarray
            The fingerprint to search for.
        k : int
            The number of neighbors to return.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.

        Returns
        -------
        dict
            (Index --> Distance)

        """
        distances = self.distances(fingerprint, mode)
        indices = np.argsort(distances)
        return {i.decode('utf-8'): d for i, d in zip(self._index[indices, 0][:k], distances[indices][:k])}

    def rnn(self, fingerprint, radius, mode="native"):
        """
        Radius-Nearest Neighbors

        Parameters
        ----------
        fingerprint : ndarray
            The fingerprint to search for.
        radius : float
            The maximum distance of neighbors to return.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.

        Returns
        -------
        dict
            (Index --> Distance)

        """
        distances = self.distances(fingerprint, mode)
        indices = np.argsort(distances)
        indices = indices[distances[indices] <= radius]
        return {i.decode('utf-8'): d for i, d in zip(self._index[indices, 0], distances[indices])}

    def distances(self, fingerprint, mode="native"):
        if mode == "native":
            return self.distances_py(fingerprint)
        elif mode == "cl":
            return self.distances_cl(fingerprint)
        else:
            raise ValueError("'mode' can only be 'native' or 'cl'")

    def max_memory_allocation_size(self):
        return min([d.max_mem_alloc_size for d in self.cl_context.devices])

    @timing(debug=True)
    def distances_cl(self, fingerprint):
        database_buffer = self.input_buffer(self.features)

        start_positions_buffer = self.input_buffer(self.start_positions)

        features_lengths_buffers = self.input_buffer(self.features_lengths)

        fingerprint_buffer = self.input_buffer(fingerprint)

        fingerprint_length = np.array([len(fingerprint)])
        fingerprint_length_buffer = self.input_buffer(fingerprint_length)

        distances = np.zeros(len(self._index), dtype=np.float32)
        distances_buf = self.output_buffer(distances)

        # Run the kernel
        queue = self.queue()
        exec_evt = self.run_kernel(queue, distances.shape, distances_buf, fingerprint_buffer, fingerprint_length_buffer,
                                   database_buffer, start_positions_buffer, features_lengths_buffers)

        # Read the result
        cl.enqueue_copy(queue, distances, distances_buf, is_blocking=True, wait_for=[exec_evt]).wait()
        queue.finish()
        return distances

    @timing(debug=True)
    def input_buffer(self, array):
        return cl.Buffer(self._ctx, MF.READ_ONLY | MF.COPY_HOST_PTR, hostbuf=array)

    @timing(debug=True)
    def output_buffer(self, array):
        return cl.Buffer(self._ctx, MF.WRITE_ONLY, size=array.nbytes)

    @timing(debug=True)
    def run_kernel(self, queue, shape, *args):
        exec_evt = self.program.nn(queue, shape, None, *args)
        return exec_evt

    @property
    def index(self):
        return self._index

    @property
    def data_frame(self):
        """
        Create a DataFrame with this model.

        Returns
        -------
        pandas.DataFrame
            A data frame.
        """
        df = DataFrame([self[i] for i in range(len(self))],
                       index=["fingerprint"], columns=self._index)
        return df.T

    def __repr__(self):
        return "NearestNeighbors %i rows" % (len(self))

    def __len__(self):
        return len(self._index)


class DBNearestNeighbors(object):
    def __init__(self, index, session, fingerprint_format, metric='jaccard'):
        self._index = index
        self._session = session
        self.fingerprint_format = fingerprint_format
        self._neighbors = None
        self._metric = metric

    @property
    def neighbors(self):
        if self._neighbors is None:
            logger.info("db-nn: building NearestNeighbors")
            self._neighbors = neighbors.NearestNeighbors(algorithm='brute', metric=self._metric)
            self._neighbors.fit(self.features)
        return self._neighbors

    def __getitem__(self, index):
        key = self._index[index]
        metabolite = self._session.query(Metabolite).filter(Metabolite.inchi_key == key).one()
        return metabolite.fingerprings[self.fingerprint_format]

    def knn(self, fingerprint, k, mode="native"):
        """
        K-Nearest Neighbors

        Parameters
        ----------

        fingerprint : ndarray
            The fingerprint to search for.
        k : int
            The number of neighbors to return.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.

        Returns
        -------
        dict
            (Index --> Distance)

        """
        logger.info("db-nn: searching for k-nearest-neighbors (%i)" % k)
        fingerprint = fingerprint.reshape(1, -1)
        logger.debug("Reshaped fingerprint %s" % fingerprint)
        distances, indices = self.neighbors.kneighbors(fingerprint, k, True)
        distances, indices = distances[0], indices[0]
        return {self.index[i]: d for i, d in zip(indices, distances)}

    def rnn(self, fingerprint, radius, mode="native"):
        """
        Radius-Nearest Neighbors

        Parameters
        ----------
        fingerprint : ndarray
            The fingerprint to search for.
        radius : float
            The maximum distance of neighbors to return.
        mode : str
            'native' to run python implementation or 'cl' to run OpenCL implementation if available.

        Returns
        -------
        dict
            (Index --> Distance)

        """
        logger.info("db-nn: searching for radius-nearest-neighbors (%.4f)" % radius)
        fingerprint = fingerprint.reshape(1, -1)
        logger.debug("Reshaped fingerprint %s" % fingerprint)
        distances, indices = self.neighbors.radius_neighbors(fingerprint, radius, True)
        distances, indices = distances[0], indices[0]
        return {self.index[i]: d for i, d in zip(indices, distances)}

    def distances(self, fingerprint, mode="native"):
        logger.info("db-nn: calculating all distances")
        fingerprint = fingerprint.reshape(1, -1)
        distances, indices = self.neighbors.radius_neighbors(fingerprint, 0, True)
        distances, indices = distances[0], indices[0]
        return {self.index[i]: d for i, d in zip(indices, distances)}

    @property
    def index(self):
        return [b.decode() for b in self._index[:, 0]]

    @property
    def features(self):
        features = [None for _ in self.index]
        query = self._session.query(Metabolite).filter(
            Metabolite.inchi_key.in_(self.index)
        ).options(load_only('id', 'inchi_key'))

        indices = {inchi_key: i for i, inchi_key in enumerate(self.index)}

        for metabolite in query.yield_per(1000):
            fp = metabolite.fingerprints[self.fingerprint_format]
            i = indices[metabolite.inchi_key]
            features[i] = fp

        assert all(f is not None for f in features)

        return features

    @property
    def data_frame(self):
        """
        Create a DataFrame with this model.

        Returns
        -------
        pandas.DataFrame
            A data frame.
        """
        df = DataFrame([self[i] for i in range(len(self))],
                       index=["fingerprint"], columns=self._index)
        return df.T

    def __repr__(self):
        return "DBNearestNeighbors %i rows" % (len(self))

    def __len__(self):
        return len(self._index)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import numpy as np
import pybel
from bitarray import bitarray

from marsi.chemistry.common import inchi_key_lru_cache

from cachetools import cached, LRUCache
from marsi.chemistry.common import convex_hull_volume, monte_carlo_volume as mc_vol


lru_cache = LRUCache(maxsize=256)


__all__ = ['has_radical', 'mol_to_inchi', 'mol_to_inchi_key', 'mol_to_svg', 'mol_chebi_id', 'mol_drugbank_id',
           'mol_pubchem_id', 'mol_str_to_inchi', 'align_molecules', 'inchi_to_molecule', 'smiles_to_molecule',
           'fingerprint', 'fingerprint_to_bits', 'get_spectrophore_data', 'inchi_to_inchi_key', 'solubility']

fps = pybel.fps


fp_bits = {
    'maccs': 167,
    'fp2': 1024,
}


def has_radical(mol):
    """
    Finds if a pybel.Molecule has Radicals.
    Radicals have an atomic number of 0.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    bool
        True if there are any radicals.
    """
    return any(a.atomicnum == 0 for a in mol.atoms)


def mol_to_inchi(mol):
    """
    Makes an InChI from a pybel.Molecule.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    str
        A InChI string.
    """
    return mol.write(format="inchi", opt=dict(errorlevel=0)).strip()


def mol_to_svg(mol):
    """
    Makes an SVG from a pybel.Molecule.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    str
        A SVG string.
    """
    return mol.write(format="svg", opt=dict(errorlevel=0)).strip()


def mol_to_inchi_key(mol):
    """
    Makes an InChI Key from a pybel.Molecule.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    str
        A InChI key.
    """
    return mol.write(format="inchikey", opt=dict(errorlevel=0)).strip()


@cached(inchi_key_lru_cache)
def inchi_to_inchi_key(inchi):
    """
    Makes an InChI Key from a InChI string.

    Parameters
    ----------
    inchi : str
        A valid InChI string.

    Returns
    -------
    str
        A InChI key.
    """
    return mol_to_inchi_key(inchi_to_molecule(inchi))


def mol_drugbank_id(mol):
    """
    Returns the DrugBank ID from the molecule data.

    Parameters
    ----------
    mo l: pybel.Molecule
        A molecule.

    Returns
    -------
    str
       DrugBank ID
    """
    return mol.data['DRUGBANK_ID'].strip()


def mol_pubchem_id(mol):
    """
    Returns the PubChem Compound ID from the molecule data.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    str
       PubChem Compound ID
    """
    return mol.data['PUBCHEM_COMPOUND_CID'].strip()


def mol_chebi_id(mol):
    """
    Returns the ChEBI ID from the molecule data.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.

    Returns
    -------
    str
       ChEBI ID
    """
    return mol.data['ChEBI ID'].strip()


def fingerprint(mol, fpformat='maccs'):
    """
    Returns the Fingerprint of the molecule.

    Parameters
    ----------
    mol : pybel.Molecule
        A molecule.
    fpformat : str
        A valid fingerprint format (see pybel.fps)

    Returns
    -------
    pybel.Fingerprint
        A fingerprint
    """
    if fpformat not in pybel.fps:
        raise AssertionError("'%s' is not a valid fingerprint format" % fpformat)
    return mol.calcfp(fptype=fpformat)


@cached(lru_cache)
def inchi_to_molecule(inchi):
    """
    Returns a molecule from a InChI string.

    Parameters
    ----------
    inchi : str
        A valid string.

    Returns
    -------
    mol : pybel.Molecule
        A molecule.
    """
    mol = pybel.readstring('inchi', inchi, opt=dict(errorlevel=0))
    mol.OBMol.StripSalts()
    mol.OBMol.Kekulize()
    _set_zero_charge(mol)
    mol.addh()

    return mol


def _set_zero_charge(mol):
    for a in mol.atoms:
        a.OBAtom.SetFormalCharge(0)


def sdf_to_molecule(from_file_or_molecule_desc, from_file=True):
    """
    Returns a molecule from a SDF file.

    Parameters
    ----------
    from_file_or_molecule_desc : str
        A valid SDF file path or a valid SDF string.
    from_file : bool
        If True tries to read the molecule from a file.


    Returns
    -------
    mol : pybel.Molecule
        A molecule.
    """
    if from_file:
        mol = next(pybel.readfile('sdf', from_file_or_molecule_desc, opt=dict(errorlevel=0)))
    else:
        mol = pybel.readstring('sdf', from_file_or_molecule_desc, opt=dict(errorlevel=0))
    mol.OBMol.StripSalts()
    mol.OBMol.Kekulize()
    _set_zero_charge(mol)
    mol.addh()
    return mol


def molecule_to_sdf(molecule):
    """
    Makes an SDF from a pybel.Molecule.

    Parameters
    ----------
    molecule : pybel.Molecule
        A molecule.

    Returns
    -------
    str
    A SDF string.
    """
    return molecule.write('sdf', opt=dict(errorlevel=0)).strip()


def mol_to_molecule(file_or_molecule_desc, from_file=True):
    """
    Returns a molecule from a MOL file.

    Parameters
    ----------
    file_or_molecule_desc : str
        A valid MOL file path or a valid MOL string.
    from_file : bool
        If True tries to read the molecule from a file.

    Returns
    -------
    mol: pybel.Molecule
        A molecule.
    """
    if from_file:
        mol = next(pybel.readfile('mol', file_or_molecule_desc, opt=dict(errorlevel=0)))
    else:
        mol = pybel.readstring('mol', file_or_molecule_desc, opt=dict(errorlevel=0))
    mol.OBMol.StripSalts()
    mol.OBMol.Kekulize()
    _set_zero_charge(mol)
    mol.addh()
    return mol


def mol_str_to_inchi(mol_str):
    """
    Returns the InChI from the molecule data.

    Parameters
    ----------
    mol_str : str
        A valid MOL string.

    Returns
    -------
    str
        A InChI string.
    """
    mol_to_inchi(pybel.readstring('mol', mol_str))


def smiles_to_molecule(smiles):
    """
    Returns the pybel.Molecule from the molecule data.

    Parameters
    ----------
    smiles : str
        A valid SMILES string.

    Returns
    -------
    pybel.Molecule
        A Molecule.
    """
    return pybel.readstring('smi', smiles)


def fingerprint_to_bits(fp, bits=1024):
    """
    Converts a pybel.Fingerprint into a binary array

    Parameters
    ----------
    fp : pybel.Fingerprint
        A fingerprint molecule.

    bits : int
        Number of bits (default is 1024)
    Returns
    -------
    bitarray
        An array of 0's and 1's.
    """
    bits_list = bitarray(bits)
    bits_list.setall(0)
    for i in fp.bits:
        if i <= bits:
            bits_list[i - 1] = 1

    return bits_list


def align_molecules(reference, molecule, include_h=True, symmetry=True):
    """
    Align molecule to a reference.

    Parameters
    ----------
    reference : pybel.Molecule
        A reference molecule.
    molecule : pybel.Molecule
        Molecule to align.
    include_h : bool
        Include implicit hydrogen atoms.
    symmetry : bool

    Returns
    -------
    list
    """

    align = pybel.ob.OBAlign
    return align(reference, molecule, include_h, symmetry)


def get_spectrophore_data(molecule):
    """
    A Spectrophore is calculated as a vector of 48 numbers (in the case of a non-stereospecific Spectrophore.
    The 48 doubles are organised into 4 sets of 12 doubles each:

    * numbers 01-11: Spectrophore values calculated from the atomic partial charges;
    * numbers 13-24: Spectrophore values calculated from the atomic lipophilicity properties;
    * numbers 25-36: Spectrophore values calculated from the atomic shape deviations;
    * numbers 37-48: Spectrophore values calculated from the atomic electrophilicity properties;

    Parameters
    ----------
    molecule : pybel.Molecule

    Returns
    -------
    ndarray:
        Float 1D-Array with the 48 features generated by OBSpectrophore.
    """
    spectrophore = pybel.ob.OBSpectrophore()

    return np.array(spectrophore.GetSpectrophore(molecule.OBMol), dtype=np.float32)


def solubility(molecule, log_value=True):
    """
    ESOL:  Estimating Aqueous Solubility Directly from Molecular Structure [1]

        $Log(S_w) = 0.16 - 0.63 logP - 0.0062 MWT + 0.066 RB - 0.74 AP$
        MWT = Molecular Weight
        RB = Rotatable Bounds
        AP = Aromatic Proportion
        ogP$

    Parameters
    ----------
    molecule : pybel.Molecule
         A molecule.
    log_value : bool
        Return log(Solubility) if true (default).

    Returns
    -------
    float
        log(S_w): log value of solubility

    """

    descriptors = molecule.calcdesc(['logP', 'MW', 'atoms'])
    log_p = descriptors['logP']
    mwt = descriptors['MW']
    rb = molecule.OBMol.NumRotors()
    atoms = molecule.atoms
    ap = len([a for a in atoms if a.OBAtom.IsAromatic()]) / descriptors['atoms']

    log_sw = 0.16 - 0.62 * log_p - 0.0062 * mwt + 0.066 * rb - 0.74 * ap
    if log_value:
        return log_sw
    else:
        return np.exp(log_sw)


def molecule_convex_hull_volume(molecule, forcefield='mmff94', steps=100):
    """
    Calculates the volume of the convex hull formed by the molecule.

    Parameters
    ----------
    molecule : pybel.Molecule
    """
    assert isinstance(molecule, pybel.Molecule)
    assert forcefield in pybel.forcefields
    molecule.addh()
    if molecule.dim < 3:
        molecule.make3D(forcefield=forcefield, steps=steps)

    xyz = [a.coords for a in molecule.atoms]
    return convex_hull_volume(xyz)


def monte_carlo_volume(molecule, coordinates=None, tolerance=1, max_iterations=10000, step_size=1000,
                       seed=time.time(), verbose=False, forcefield='mmff94', steps=100):
    """
    Adapted from:

    Simple Monte Carlo estimation of VdW molecular volume (in A^3)
    by Geoffrey Hutchison <geoffh@pitt.edu>

    https://github.com/ghutchis/hutchison-cluster

    Parameters
    ----------
    molecule : pybel.Molecule
        A molecule from pybel.
    coordinates : list
        A list of pre selected x,y,z coords. It must match the atoms order.
    tolerance : float
        The tolerance for convergence of the monte carlo abs(new_volume - volume) < tolerance
    max_iterations : int
        Number of iterations before the algorithm starts
    step_size : int
        Number of points to add each step.
    seed : object
        A valid seed for random.
    verbose : bool
        Print debug information if True.
    forcefield : str
        The force field to get a 3D molecule. (only if it is not 3D already)
    steps : int
        The number of steps used for the force field to get a 3D molecule. (only if it is not 3D already)

    Returns
    -------
    float
        Molecule volume.
    """

    assert isinstance(molecule, pybel.Molecule)
    if coordinates is None:
        if molecule.dim < 3:
            molecule.make3D(forcefield=forcefield, steps=steps)
        coordinates = np.array([a.coords for a in molecule.atoms], dtype=np.float32)
    else:
        coordinates = np.array(coordinates, dtype=np.float32)
    vdw_radii = np.array([pybel.ob.etab.GetVdwRad(a.atomicnum) for a in molecule.atoms], dtype=np.float32)
    return mc_vol(coordinates, vdw_radii, tolerance, max_iterations, step_size, seed, int(verbose))



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pandas import DataFrame

from marsi.utils import data_dir

__all__ = ["chebi", "drugbank"]

chebi = DataFrame.from_csv(os.path.join(data_dir, "chebi_analogues_filtered.csv"))

drugbank = DataFrame.from_csv(os.path.join(data_dir, "drugbank_open_vocabulary.csv"), sep=",", index_col=None)
drugbank.columns = ["id", "accessions", "common_names", "cas", "unii", "synonyms", "inchi_key"]

drugbank.fillna("", inplace=True)

drugbank['accessions'] = drugbank.accessions.apply(str.split, args=(" | ",))
drugbank['synonyms'] = drugbank.synonyms.apply(str.split, args=(" | ",))

pubchem = DataFrame.from_csv(os.path.join(data_dir, "pubchem_data.csv"))

kegg = DataFrame.from_csv(os.path.join(data_dir, "kegg_data.csv"))

# binding_db = read_csv(os.path.join(data_dir, "BindingDB_All.tsv"), sep="\t", error_bad_lines=False)
# with open(os.path.join(data_dir, 'binding_db_sane_columns.txt'), 'r') as columns_file:
#     line = next(columns_file)
#     binding_db.columns = line.split(", ")

# solubility = DataFrame.from_csv(os.path.join(data_dir, 'solubility.csv'))
# solubility.columns = ['log_measured', 'log_predicted', 'smiles']



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from cement.core.controller import CementBaseController, expose

from marsi.config import prj_dir, db_name
from marsi.utils import data_dir, log_dir, models_dir


class MarsiBaseController(CementBaseController):
    """
    This is the application base controller.

    """

    class Meta:
        label = 'base'

    @expose(hide=True)
    def default(self):
        """
        1. Download necessary files.
        2. Build the data files.
        3. Build the database.

        Returns
        -------

        """
        if not os.path.isdir(data_dir):
            os.mkdir(data_dir)
        if not os.path.isdir(log_dir):
            os.mkdir(log_dir)
        if not os.path.isdir(models_dir):
            os.mkdir(models_dir)
        print("\n################################################################")
        print("#                                                              #")
        print("#                       Welcome to MARSI!                      #")
        print("#                                                              #")
        print("################################################################")

        print("Your current project directory is %s" % prj_dir)
        print("The directory tree is:")
        print("%s:" % prj_dir)
        print("-- Data files: %s" % data_dir.split(os.sep)[-1])
        print("-- Log files:  %s" % log_dir.split(os.sep)[-1])
        print("-- Models dir: %s" % models_dir.split(os.sep)[-1])
        print("Your current database is %s" % db_name)
        print("\n")
        print("If you want to change this settings you can create a 'setup.cfg' as follows:\n")
        print("######### Begin MARSI configuration #######")
        print("[marsi]                                    ")
        print("# Working directory                        ")
        print("prj_dir=<directory of your choice>         ")
        print("# Name of the database                     ")
        print("db_name='marsi-db'                         ")
        print("######### End of MARSI configuration ######\n")
        print("")
        print("For more information about database configuration try: https://biosustain.github.io/marsi")

    @expose(help="Show db status")
    def db_status(self):
        from marsi.io.db import Database

        print("Database status (%s):" % db_name)
        print("+----------------------------------------------+--------------+")
        print("| Collection                                   | Total        |")
        print("+----------------------------------------------+--------------+")
        print("| Chemical compounds                           | %s |" % str(len(Database.metabolites)).ljust(12))
        print("+----------------------------------------------+--------------+")



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

import inspyred
import numpy
from IProgress import ProgressBar, Bar, Percentage
from cameo.core.strain_design import StrainDesignMethod, StrainDesign, StrainDesignMethodResult
from cameo.core.utils import get_reaction_for
from cameo.flux_analysis.analysis import phenotypic_phase_plane, flux_variability_analysis
from cameo.flux_analysis.simulation import fba
from cameo.strain_design.heuristic.evolutionary.archives import ProductionStrainArchive
from cameo.strain_design.heuristic.evolutionary.objective_functions import biomass_product_coupled_min_yield, \
    biomass_product_coupled_yield
from cameo.visualization.plotting import plotter
from cobra.core.model import Model
from cobra.exceptions import OptimizationError
from pandas import DataFrame

from marsi.cobra.flux_analysis.manipulation import knockout_metabolite
from marsi.cobra.strain_design.metaheuristic import MetaboliteKnockoutOptimization
from marsi.cobra.strain_design.target import MetaboliteKnockoutTarget
from marsi.utils import search_metabolites

logger = logging.getLogger(__name__)

__all__ = ["OptMet"]


class OptMet(StrainDesignMethod):

    def __init__(self, model, evolutionary_algorithm=inspyred.ec.GA,
                 essential_metabolites=None, plot=True, *args, **kwargs):
        super(OptMet, self).__init__(*args, **kwargs)
        self._model = model
        self._algorithm = evolutionary_algorithm
        self._optimization_algorithm = None
        self._essential_metabolites = essential_metabolites
        self._plot = plot
        self._manipulation_type = "metabolites"

    @property
    def manipulation_type(self):
        return self._manipulation_type

    @property
    def plot(self):
        return self._plot

    @plot.setter
    def plot(self, plot):
        self._plot = plot
        if self._optimization_algorithm is not None:
            self._optimization_algorithm.plot = plot

    def run(self, target=None, biomass=None, substrate=None, max_knockouts=5, variable_size=True,
            simulation_method=fba, growth_coupled=False, max_evaluations=20000, population_size=200,
            max_results=50, seed=None, **kwargs):
        """
        Parameters
        ----------
        target : str, Metabolite or Reaction
            The design target
        biomass : str, Metabolite or Reaction
            The biomass definition in the model
        substrate : str, Metabolite or Reaction
            The main carbon source
        max_knockouts : int
            Max number of knockouts allowed
        variable_size : bool
            If true, all candidates have the same size. Otherwise the candidate size can be from 1 to max_knockouts.
        simulation_method: function
            Any method from cameo.flux_analysis.simulation or equivalent
        growth_coupled : bool
            If true will use the minimum flux rate to compute the fitness
        max_evaluations : int
            Number of evaluations before stop
        population_size : int
            Number of individuals in each generation
        max_results : int
            Max number of different designs to return if found.
        kwargs : dict
            Arguments for the simulation method.
        seed : int
            A seed for random.

        Returns
        -------
        OptMetResult
        """

        target = get_reaction_for(self._model, target)
        biomass = get_reaction_for(self._model, biomass)
        substrate = get_reaction_for(self._model, substrate)

        if growth_coupled:
            objective_function = biomass_product_coupled_min_yield(biomass, target, substrate)
        else:
            objective_function = biomass_product_coupled_yield(biomass, target, substrate)

        optimization_algorithm = MetaboliteKnockoutOptimization(
            model=self._model,
            heuristic_method=self._algorithm,
            essential_metabolites=self._essential_metabolites,
            objective_function=objective_function,
            plot=self.plot)

        optimization_algorithm.simulation_kwargs = kwargs
        optimization_algorithm.simulation_method = simulation_method
        optimization_algorithm.archiver = ProductionStrainArchive()

        result = optimization_algorithm.run(max_evaluations=max_evaluations,
                                            pop_size=population_size,
                                            max_size=max_knockouts,
                                            variable_size=variable_size,
                                            maximize=True,
                                            max_archive_size=max_results,
                                            seed=seed,
                                            **kwargs)

        kwargs.update(optimization_algorithm.simulation_kwargs)
        return OptMetResult(self._model, result, objective_function, simulation_method, self._manipulation_type,
                            biomass, target, substrate, kwargs)


class OptMetResult(StrainDesignMethodResult):
    __method_name__ = "OptMet"

    def __init__(self, model, knockouts, objective_function, simulation_method, manipulation_type, biomass, target,
                 substrate, simulation_kwargs, *args, **kwargs):

        assert isinstance(model, Model)

        self._model = model
        self._knockouts = knockouts
        self._objective_function = objective_function
        self._simulation_method = simulation_method
        self._manipulation_type = manipulation_type
        self._biomass = biomass
        self._target = target
        self._substrate = substrate
        self._processed_solutions = None
        self._simulation_kwargs = simulation_kwargs

        super(OptMetResult, self).__init__(self.designs, *args, **kwargs)

    @property
    def designs(self):
        if self._processed_solutions is None:
            self._process_solutions()
        return self._processed_solutions.designs.tolist()

    def _repr_html_(self):
        return """
        <h3>OptMet Result</h3>
        <ul>
            <li>Simulation: %s<br/></li>
            <li>Objective Function: %s<br/></li>
        </ul>
        %s
        """ % (self._simulation_method.__name__,
               self._objective_function._repr_latex_(),
               self.data_frame._repr_html_())

    @property
    def data_frame(self):
        if self._processed_solutions is None:
            self._process_solutions()

        return DataFrame(self._processed_solutions)

    def _process_solutions(self):
        processed_solutions = DataFrame(columns=["designs", "size", "fva_min", "fva_max",
                                                 "target_flux", "biomass_flux", "yield", "fitness"])

        if len(self._knockouts) == 0:
            logger.warning("No solutions found")
            self._processed_solutions = processed_solutions

        else:
            progress = ProgressBar(maxval=len(self._knockouts), widgets=["Processing solutions: ", Bar(), Percentage()])
            for i, solution in progress(enumerate(self._knockouts)):
                try:
                    processed_solutions.loc[i] = process_metabolite_knockout_solution(
                        self._model, solution, self._simulation_method, self._simulation_kwargs,
                        self._biomass, self._target, self._substrate, self._objective_function)
                except OptimizationError as e:
                    logger.error(e)
                    processed_solutions.loc[i] = [numpy.nan for _ in processed_solutions.columns]

            self._processed_solutions = processed_solutions

    def display_on_map(self, index=0, map_name=None, palette="YlGnBu"):
        with self._model:
            for ko in self.data_frame.loc[index, "metabolites"]:
                knockout_metabolite(self._model, self._model.metabolites.get_by_id(ko))
            fluxes = self._simulation_method(self._model, **self._simulation_kwargs)
            fluxes.display_on_map(map_name=map_name, palette=palette)

    def plot(self, index=0, grid=None, width=None, height=None, title=None, palette=None, **kwargs):
        wt_production = phenotypic_phase_plane(self._model, objective=self._target, variables=[self._biomass])
        with self._model:
            for ko in self.data_frame.loc[index, "metabolites"]:
                knockout_metabolite(self._model, self._model.metabolites.get_by_id(ko))
            mt_production = phenotypic_phase_plane(self._model, objective=self._target, variables=[self._biomass])

        if title is None:
            title = "Production Envelope"

        data_frame = DataFrame(columns=["ub", "lb", "value", "strain"])
        for _, row in wt_production.iterrows():
            _df = DataFrame([[row['objective_upper_bound'], row['objective_lower_bound'], row[self._biomass.id], "WT"]],
                            columns=data_frame.columns)
            data_frame = data_frame.append(_df)
        for _, row in mt_production.iterrows():
            _df = DataFrame([[row['objective_upper_bound'], row['objective_lower_bound'], row[self._biomass.id], "MT"]],
                            columns=data_frame.columns)
            data_frame = data_frame.append(_df)

        plot = plotter.production_envelope(data_frame, grid=grid, width=width, height=height, title=title,
                                           x_axis_label=self._biomass.id, y_axis_label=self._target.id, palette=palette)
        plotter.display(plot)


def process_metabolite_knockout_solution(model, solution, simulation_method, simulation_kwargs, biomass, target,
                                         substrate, objective_function):
    """
    Parameters
    ----------

    model : SolverBasedModel
        A constraint-based model
    solution : tuple
        The output of a decoder (metabolite species)
    simulation_method : function
        See see cameo.flux_analysis.simulation
    simulation_kwargs : dict
        Keyword arguments to run the simulation method
    biomass : Reaction
        Cellular biomass reaction
    target : Reaction
        The strain design target
    substrate : Reaction
        The main carbon source uptake rate
    objective_function : function
        A cameo.strain_design.heuristic.evolutionary.objective_functions.ObjectiveFunction

    Returns
    -------

    list
        A list with: metabolites, size, fva_min, fva_max, target flux, biomass flux, yield, fitness
    """

    metabolite_targets = [search_metabolites(model, species_id) for species_id in solution]
    with model:
        for metabolites in metabolite_targets:
            for metabolite in metabolites:
                knockout_metabolite(model, metabolite)

        reactions = objective_function.reactions
        flux_dist = simulation_method(model, reactions=reactions, objective=biomass, **simulation_kwargs)
        model.objective = biomass

        fva = flux_variability_analysis(model, fraction_of_optimum=0.99, reactions=[target])
        target_yield = flux_dist[target] / abs(flux_dist[substrate])

        fitness = objective_function(model, flux_dist, solution)

        fva_min, fva_max = fva.lower_bound(target), fva.upper_bound(target)
        target_flux, biomass_flux = flux_dist[target], flux_dist[biomass]

        design = StrainDesign(MetaboliteKnockoutTarget(species_id=species_id) for species_id in solution)

        return design, len(design), fva_min, fva_max, target_flux, biomass_flux, target_yield, fitness



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function

import numpy
from Cython.Build import cythonize
from setuptools import setup, find_packages, Extension

import versioneer

requirements = ['pandas>=0.18.1',
                'numpy>=1.11.1',
                'bioservices>=1.4.14',
                'requests>=2.11.1',
                'bokeh>=0.12',
                'cameo>=0.11',
                'sqlalchemy>=1.1.9',
                'scikit-learn>=0.18.1',
                'psycopg2>=2.7.1',
                'bitarray>=0.8.1',
                'Cython>=0.24',
                'cement>2.10',
                'pubchempy>=1.0.3 ',
                'cachetools>=2.0.0',
                'alembic>=0.9']

extra_requirements = {
    'docs': ['Sphinx>=1.3.5', 'numpydoc>=0.5'],
    'jupyter': ['jupyter>=1.0.0', 'ipywidgets>=4.1.1'],
    'test': ['pytest>=1.3.7', 'pytest-cov>=2.4', 'pytest-benchmark>=3.0'],
    '3d': ['imolecule>=0.1.13'],
    'opencl': ['pyopencl>=2016.1']
}

extra_requirements['all'] = sum([list(values) for values in extra_requirements.values()], [])


ext_modules = cythonize([Extension("marsi.chemistry.common_ext", 
                                   sources=["marsi/chemistry/common_ext.pyx"],
                                   include_dirs=[numpy.get_include()]), 
                         Extension("marsi.nearest_neighbors.model_ext",
                                   sources=["marsi/nearest_neighbors/model_ext.pyx"],
                                   include_dirs=[numpy.get_include()])
                        ])

include_dirs = [numpy.get_include()]


setup(
    name='marsi',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(exclude=("*.c",)),
    package_data={'marsi': ['alembic.ini', 'alembic/*',
                            'chemistry/common_ext.pxd', 'chemistry/common_ext.pyx',
                            'nearest_neighbors/model_ext.pyx']},
    install_requires=requirements,
    extras_require=extra_requirements,
    ext_modules=ext_modules,
    scripts=['bin/marsi'],
    include_package_data=True,
    author='Joao Cardoso',
    author_email='joaca@biosustain.dtu.dk',
    description='marsi - Metabolite Analogues for Rational Strain Improvement',
    license='Apache License Version 2.0',
    keywords='biology metabolism bioinformatics chemoinformatics',
    url='https://github.com/biosustain/marsi',
    long_description=open("README.md").read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.4',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

    ],
    include_dirs=include_dirs
)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from collections import Counter

from IProgress.progressbar import ProgressBar
from IProgress.widgets import Percentage, Bar, ETA
from cachetools import cached, LRUCache
from cameo.flux_analysis.analysis import find_essential_metabolites

from marsi import bigg_api
from marsi.io.bigg import bigg_metabolites
from marsi.io.enrichment import inchi_from_chebi, inchi_from_kegg


__all__ = ['find_inchi_for_bigg_metabolite', 'annotate_metabolite', 'annotate_model']


lru_cache = LRUCache(maxsize=1024)

logger = logging.getLogger(__name__)

CURRENCY_METABOLITES = {"atp", "adp", "nad", "nadh", "nadp", "nadph", "amp",
                        "h2o", "h", "coa", "acp", "pi", 'pppi', 'ppi'}

DATABASE_LINKS = 'database_links'

CHEBI = 'CHEBI'
KEGG = "KEGG Compound"


@cached(lru_cache)
def find_inchi_for_bigg_metabolite(model_id, metabolite_id):
    try:
        links = bigg_metabolites.loc[metabolite_id].database_links
    except KeyError:
        metabolite_data = bigg_api.get_model_metabolite(model_id, metabolite_id)
        links = metabolite_data[DATABASE_LINKS]
    inchi_keys = []
    if CHEBI in links:
        inchi_keys += [inchi_from_chebi(link['id']) for link in links[CHEBI]]

    if KEGG in links:
        inchi_keys += [inchi_from_kegg(link['id']) for link in links[KEGG]]

    counter = Counter(inchi_keys)
    del counter[None]

    if len(counter) > 0:
        return max(counter, key=lambda key: counter[key])
    else:
        raise ValueError(metabolite_id)


def annotate_metabolite(metabolite):
    try:
        metabolite.annotation['inchi']
    except KeyError:
        try:
            metabolite.annotation['inchi'] = find_inchi_for_bigg_metabolite(metabolite.model, metabolite.id)
        except ValueError:
            pass


def annotate_model(model):
    pbar = ProgressBar(maxval=len(model.metabolites), widgets=["Annotating: ", Percentage(), Bar(), ETA()])
    [annotate_metabolite(m) for m in pbar(model.metabolites)]


def essential_species_ids(model):
    essential_metabolites = find_essential_metabolites(model)
    return {m.id[:-2] for m in essential_metabolites}



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython
cimport cython

import numpy as np
cimport numpy as np

from numpy import ndarray

from libc.math cimport sqrt

cdef extern int __builtin_popcount(unsigned int) nogil

@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef float _tanimoto_coefficient(np.ndarray[INT32_t, ndim=1] fingerprint1, np.ndarray[INT32_t, ndim=1] fingerprint2):
    """
    Calculate the Tanimoto coefficient for 2 fingerprints.

    Parameters
    ----------
    fingerprint1 : ndarray
        First fingerprint.
    fingerprint2 : ndarray
        Second fingerprint.

    Returns
    -------
    float
        The Tanimoto coefficient.
    """
    cdef unsigned int len1 = fingerprint1.shape[0]
    cdef unsigned int len2 = fingerprint2.shape[0]

    if len1 != len2:
        return -1.0

    cdef int and_bits = 0
    cdef int or_bits = 0

    cdef int fp_and
    cdef int fp_or

    for i in range(len1):
        fp_and = fingerprint1[i] & fingerprint2[i]
        fp_or = fingerprint1[i] | fingerprint2[i]

        and_bits += __builtin_popcount(fp_and)
        or_bits += __builtin_popcount(fp_or)

    return float(and_bits)/float(or_bits)

@cython.nonecheck(False)
@cython.cdivision(True)
cdef float _tanimoto_distance(np.ndarray[INT32_t, ndim=1] fingerprint1, np.ndarray[INT32_t, ndim=1] fingerprint2):
    """
    Calculate the Tanimoto distance for 2 fingerprints (1 - tanimoto coefficient).

     Arguments
    ---------
    fingerprint1: ndarray
        First fingerprint.
    fingerprint2: ndarray
        Second fingerprint.

    Returns
    -------
    float
        The Tanimoto distance.
    """

    return 1 - _tanimoto_coefficient(fingerprint1, fingerprint2)


def tanimoto_coefficient(np.ndarray[INT32_t, ndim=1] fingerprint1, np.ndarray[INT32_t, ndim=1] fingerprint2):
    """
    Calculate the Tanimoto coefficient for 2 fingerprints.

    Parameters
    ----------
    fingerprint1 : ndarray
        First fingerprint.
    fingerprint2 : ndarray
        Second fingerprint.

    Returns
    -------
    float
        The Tanimoto coefficient.
    """
    return _tanimoto_coefficient(fingerprint1, fingerprint2)


def tanimoto_distance(np.ndarray[INT32_t, ndim=1] fingerprint1, np.ndarray[INT32_t, ndim=1] fingerprint2):
    """
    Calculate the Tanimoto distance for 2 fingerprints (1 - tanimoto coefficient).

    Parameters
    ----------
    fingerprint1 : ndarray
        First fingerprint.
    fingerprint2 : ndarray
        Second fingerprint.

    Returns
    -------
    float
        The Tanimoto distance.
    """

    return _tanimoto_distance(fingerprint1, fingerprint2)


def rmsd(np.ndarray[FLOAT32_t, ndim=3] v, np.ndarray[FLOAT32_t, ndim=3] w):
    """
    Root-mean-squared deviation of XYZ.

    $$RMSD = \sqrt{\frac{1}{n} \sum_{i=1}^{n}{(v - t)^2}}}$$

    Parameters
    ----------
    v : list
        List of x, y, z
    w : list
        List of x, y, z
    """

    if len(v) != len(w):
        raise ValueError("Length of vectors should be the same")

    cdef unsigned int n = len(v)
    cdef unsigned int sum = 0

    for i in range(n):
        sum += (v[i] - w[i])**2 + (v[i] - w[i])**2 + (v[i] - w[i])**2

    return sqrt(1/n * sum)

@cython.nonecheck(False)
@cython.cdivision(True)
cdef limits(np.ndarray[FLOAT32_t, ndim=2] coords, np.ndarray[FLOAT32_t, ndim=1] vdw_radii, int index):
    cdef unsigned int i = 0
    cdef unsigned int n = len(coords)
    cdef np.ndarray[FLOAT32_t, ndim=1] _coord = np.zeros(n, dtype=np.float32)

    for i in range(len(coords)):
        _coord[i] = coords[i][index] + vdw_radii[i]
    return min(_coord), max(_coord)

@cython.nonecheck(False)
@cython.cdivision(True)
cdef unsigned int point_in_molecule(np.ndarray[FLOAT32_t, ndim=2] coords, np.ndarray[FLOAT32_t, ndim=1] vdw_radii, np.ndarray[FLOAT32_t, ndim=1] point):

    cdef float r, x, y, z
    cdef unsigned int n = len(coords)
    cdef np.ndarray[FLOAT32_t, ndim=1] _coord
    for i in range(n):
        _coord = coords[i]
        r = vdw_radii[i]

        x = point[0] - _coord[0]
        y = point[1] - _coord[1]
        z = point[2] - _coord[2]

        if x*x + y*y + z*z <= r**2:
            return 1
    return 0

@cython.nonecheck(False)
@cython.cdivision(True)
cdef _monte_carlo_volume(np.ndarray[FLOAT32_t, ndim=2] coords, np.ndarray[FLOAT32_t, ndim=1] vdw_radii, float tolerance,
                         unsigned int max_iterations, unsigned int step_size, int seed, unsigned int verbose):

    np.random.seed(seed)

    # determine the bounding box
    cdef float[2] mol_x = limits(coords, vdw_radii, 0)
    cdef float[2] mol_y = limits(coords, vdw_radii, 2)
    cdef float[2] mol_z = limits(coords, vdw_radii, 2)

    cdef float mol_x_len = mol_x[1] - mol_x[0]
    cdef float mol_y_len = mol_y[1] - mol_y[0]
    cdef float mol_z_len = mol_y[1] - mol_y[0]

    cdef float mol_x_mid = (mol_x[0] + mol_x[1]) / 2.0
    cdef float mol_y_mid = (mol_y[0] + mol_y[1]) / 2.0
    cdef float mol_z_mid = (mol_z[0] + mol_z[1]) / 2.0

    cdef float box_x_min = mol_x_mid - 2.0 * mol_x_len
    cdef float box_x_max = mol_x_mid + 2.0 * mol_x_len
    cdef float box_y_min = mol_y_mid - 2.0 * mol_y_len
    cdef float box_y_max = mol_y_mid + 2.0 * mol_y_len
    cdef float box_z_min = mol_z_mid - 2.0 * mol_z_len
    cdef float box_z_max = mol_z_mid + 2.0 * mol_z_len

    cdef float box_x_len = box_x_max - box_x_min
    cdef float box_y_len = box_y_max - box_y_min
    cdef float box_z_len = box_z_max - box_z_min
    cdef float box_volume = box_x_len * box_y_len * box_z_len

    if verbose:
        print("Box volume %.5f" % box_volume)

    # calculate the ratio of points inside and outside the VdW shell of the molecule

    # one hundred thousand points to start
    cdef unsigned int total_points = 100000
    cdef unsigned int points_in_molecule = 0

    cdef np.ndarray[FLOAT32_t, ndim=2] points = np.random.uniform(box_x_min, box_x_max, (total_points, 3)).astype(np.float32)
    cdef np.ndarray[FLOAT32_t, ndim=1] point
    cdef int i =0
    for i in range(total_points):
        point = points[i]
        points_in_molecule += point_in_molecule(coords, vdw_radii, point)

    # Adding more points until reach some convergence
    cdef float volume = box_volume * float(points_in_molecule) / float(total_points)
    cdef float new_volume = 0.0
    cdef j = 0
    while abs(volume - new_volume) > tolerance and j <= max_iterations:
        points = np.random.uniform(box_x_min, box_x_max, (step_size, 3)).astype(np.float32)
        for i in range(step_size):
            point = points[i]
            points_in_molecule += point_in_molecule(coords, vdw_radii, point)
        total_points += step_size
        volume = new_volume
        new_volume =  box_volume * float(points_in_molecule) / float(total_points)
        if verbose:
            print("Iteration %i, volume: %.5f, new volume: %.5f" % (j, volume, new_volume))
        j += 1

    return volume

def monte_carlo_volume(np.ndarray[FLOAT32_t, ndim=2] coords, np.ndarray[FLOAT32_t, ndim=1] vdw_radii, float tolerance,
                         int max_iterations, int step_size, int seed, int verbose):
    """
    Adapted from:

    Simple Monte Carlo estimation of VdW molecular volume (in A^3)
    by Geoffrey Hutchison <geoffh@pitt.edu>

    https://github.com/ghutchis/hutchison-cluster
    """

    return _monte_carlo_volume(coords, vdw_radii, tolerance, max_iterations, step_size, seed, verbose)



# Copyright 2016 Chr. Hansen A/S and The Novo Nordisk Foundation Center for Biosustainability, DTU.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import multiprocessing
import os
from queue import Empty

import numpy as np
import six
from IProgress import ProgressBar, Bar, ETA
from cameo.parallel import SequentialView
from pandas import DataFrame
from sqlalchemy import and_
from sqlalchemy.orm import sessionmaker

from marsi import config
from marsi.chemistry import SOLUBILITY
from marsi.chemistry import rdkit
from marsi.chemistry.molecule import Molecule
from marsi.config import default_session, engine
from marsi.io.db import Database
from marsi.io.db import Metabolite
from marsi.nearest_neighbors.model import NearestNeighbors, DistributedNearestNeighbors, DBNearestNeighbors
from marsi.utils import data_dir, INCHI_KEY_TYPE, unpickle_large, pickle_large


__all__ = ['build_nearest_neighbors_model', 'load_nearest_neighbors_model']

MODEL_FILE = os.path.join(data_dir, "fingerprints_default_%s_sol_%s.pickle")


class FeatureReader(object):
    """
    Accessory class to build fingerprints in chunks.

    Attributes
    ----------
    db : str
        The name of the database to connect to.
    fpformat : str
        The fingerprint format (see pybel.fps).
    solubility : str
        One of 'high', 'medium', 'low', 'all'.
    connection_args : dict
        Other arguments of connect


    """
    def __init__(self, db, fpformat='maccs', solubility='high', **connection_args):
        self.db = db
        self.fpformat = fpformat
        self.solubility = solubility
        self.connection_args = connection_args

    def __call__(self, index):
        subset = Database.metabolites[index[0]:index[1]]
        indices = []
        fingerprints = []
        fingerprint_lengths = []
        for m in subset:
            if SOLUBILITY[self.solubility](m.solubility):
                fingerprint = m.fingerprint(fpformat=self.fpformat)
                fingerprints.append(fingerprint)
                indices.append(m.inchi_key)
                fingerprint_lengths.append(len(fingerprint))

        _indices = np.ndarray((len(indices), 1), dtype=INCHI_KEY_TYPE)
        for i in range(_indices.shape[0]):
            _indices[i] = indices[i]
        del indices
        return _indices, fingerprints, fingerprint_lengths


def build_feature_table(database, fpformat='ecfp10', chunk_size=None, solubility='high',
                        database_name=config.db_name, view=SequentialView()):
    reader = FeatureReader(database_name, fpformat=fpformat, solubility=solubility)
    chunk_size = math.ceil(chunk_size)
    n_chunks = math.ceil(len(database) / chunk_size)
    chunks = [((i - 1) * chunk_size, i * chunk_size) for i in range(1, n_chunks + 1)]
    res = view.map(reader, chunks)
    indices = np.ndarray((0, 1), dtype=INCHI_KEY_TYPE)
    fingerprints = []
    fingerprint_lengths = []
    for r in res:
        indices = np.concatenate([indices, r[0]])
        fingerprints += r[1]
        fingerprint_lengths += r[2]
    return indices, fingerprints, fingerprint_lengths


def _build_nearest_neighbors_model(indices, features, lengths, n_models):
    chunk_size = math.ceil(len(indices) / n_models)
    chunks = [((i - 1) * chunk_size, i * chunk_size) for i in range(1, n_models + 1)]
    models = []
    for start, end in chunks:
        models.append(NearestNeighbors(indices[start:end], features[start:end], lengths[start:end]))

    return DistributedNearestNeighbors(models)


def build_nearest_neighbors_model(database, fpformat='fp4', solubility='high', n_models=5,
                                  chunk_size=1e6, view=SequentialView):
    """
    Loads a NN model.

    If a 'default_model.pickle' exists in data it will load the model. Otherwise it will build a model from the
    Database. This can take several hours depending on the size of the database.

    Parameters
    ----------
    database : marsi.io.mongodb.CollectionWrapper
        A Database interface to the metabolites.
    chunk_size : int
        Maximum number of entries per chunk.
    fpformat : str
        The format of the fingerprint (see pybel.fps)
    solubility : str
        One of high, medium, low or all.
    view : cameo.parallel.SequentialView, cameo.parallel.MultiprocesingView
        A view to control parallelization.
    n_models : int
        The number of NearestNeighbors models.
    """

    indices, features, lens = build_feature_table(database, fpformat=fpformat, chunk_size=chunk_size,
                                                  solubility=solubility, view=view)
    return _build_nearest_neighbors_model(indices, features, lens, n_models)


def load_nearest_neighbors_model(chunk_size=1e6, fpformat="fp4", solubility='all', session=default_session,
                                 view=SequentialView(), model_size=100000, source="db", costum_query=None):
    """
    Loads a NN model.

    If a 'default_model.pickle' exists in data it will load the model. Otherwise it will build a model from the
    Database. This can take several hours depending on the size of the database.

    Parameters
    ----------
    chunk_size : int
        Maximum number of entries per chunk.
    fpformat : str
        The format of the fingerprint (see pybel.fps)
    solubility : str
        One of high, medium, low or all.
    view : cameo.parallel.SequentialView, cameo.parallel.MultiprocesingView
        A view to control parallelization.
    model_size : int
        The size of each NearestNeighbor in the ensemble.
    """

    if source == "file":
        load_nearest_neighbors_model_from_file(chunk_size=chunk_size, fpformat=fpformat, solubility=solubility,
                                               view=view, model_size=model_size)
    else:
        load_nearest_neighbors_model_from_db(fpformat=fpformat, solubility=solubility,
                                             model_size=model_size, session=session, costum_query=costum_query)


def load_nearest_neighbors_model_from_db(fpformat="fp4", solubility='all', model_size=1000, session=default_session,
                                         custom_query=None):
    """
    Loads a NN model.

    If a 'default_model.pickle' exists in data it will load the model. Otherwise it will build a model from the
    Database. This can take several hours depending on the size of the database.

    Parameters
    ----------
    fpformat : str
        The format of the fingerprint (see pybel.fps)
    solubility : str
        One of high, medium, low or all.
    model_size : int
        The size of each NearestNeighbor in the ensemble.
    session : Session
        SQLAlchemy session.
    custom_query : ClauseElement
        A query to filter elements from the database.

    """

    if custom_query is not None:
        indices = np.array(session.query(Metabolite.inchi_key).filter(custom_query).all(), dtype=INCHI_KEY_TYPE)
    else:
        indices = np.array(session.query(Metabolite.inchi_key).all(), dtype=INCHI_KEY_TYPE)

    n_models = math.ceil(len(indices) / model_size)
    chunk_size = math.ceil(len(indices) / n_models)
    chunks = [((i - 1) * chunk_size, i * chunk_size) for i in range(1, n_models + 1)]
    models = []

    for start, end in chunks:
        models.append(DBNearestNeighbors(indices[start:end], session, fpformat))

    return DistributedNearestNeighbors(models)


def load_nearest_neighbors_model_from_file(chunk_size=1e6, fpformat="fp4", solubility='all',
                                           view=SequentialView(), model_size=100000):
    """
    Loads a NN model from file.

    If a 'default_model.pickle' exists in data it will load the model. Otherwise it will build a model from the
    Database. This can take several hours depending on the size of the database.

    Parameters
    ----------
    chunk_size : int
        Maximum number of entries per chunk.
    fpformat : str
        The format of the fingerprint (see pybel.fps)
    solubility : str
        One of high, medium, low or all.
    view : cameo.parallel.SequentialView, cameo.parallel.MultiprocesingView
        A view to control parallelization.
    model_size : int
        The size of each NearestNeighbor in the ensemble.
    """
    if solubility not in SOLUBILITY:
        raise ValueError('%s not one of %s' % (solubility, ", ".join(SOLUBILITY.keys())))

    model_file = MODEL_FILE % (fpformat, solubility)
    if os.path.exists(model_file):
        _indices, _features, _lengths = unpickle_large(model_file, progress=True)
    else:
        print("Building search model (fp: %s, solubility: %s)" % (fpformat, solubility))
        _indices, _features, _lengths = build_feature_table(Database.metabolites,
                                                            chunk_size=chunk_size,
                                                            fpformat=fpformat,
                                                            solubility=solubility,
                                                            view=view)
        pickle_large((_indices, _features, _lengths), model_file, progress=True)

    n_models = math.ceil(len(_indices) / model_size)
    nn_model = _build_nearest_neighbors_model(_indices, _features, _lengths, n_models)
    return nn_model


class DataBuilder(multiprocessing.Process):
    def __init__(self, inchi, task_queue, results_queue, atoms_weight, bonds_weight, timeout,
                 *args, **kwargs):
        super(DataBuilder, self).__init__(*args, **kwargs)
        self._inchi = inchi
        self._session = None
        self._tasks = task_queue
        self._results = results_queue
        self._atoms_weight = atoms_weight
        self._bonds_weight = bonds_weight
        self._timeout = timeout

    @property
    def session(self):
        if self._session is None:
            session_maker = sessionmaker(engine)
            self._session = session_maker()
        return self._session

    @property
    def molecule(self):
        return rdkit.inchi_to_molecule(self._inchi)

    def run(self):
        while True:
            try:
                inchi_key, distance = self._tasks.get()
            except Empty:
                break
            else:
                result = self.apply_similarity(inchi_key, distance)
                self._results.put(result)

        if self._session is not None:
            self.session.close()
            self.session.bind.dispose()

    def apply_similarity(self, inchi_key, distance):
        met = Metabolite.get(inchi_key=inchi_key, session=self.session)

        try:
            molecule = met.molecule('rdkit', get3d=False)
            structural_similarity = rdkit.structural_similarity(self.molecule, molecule,
                                                                atoms_weight=self._atoms_weight,
                                                                bonds_weight=self._bonds_weight,
                                                                timeout=self._timeout)
            return [inchi_key, met.formula, met.num_atoms, met.num_bonds, 1 - distance, structural_similarity]
        except Exception as e:
            print(e)
            return None


def search_closest_compounds(molecule, nn_model=None, fp_cut=0.5, fpformat="maccs", atoms_diff=3,
                             bonds_diff=3, rings_diff=2, session=default_session,
                             atoms_weight=0.5, bonds_weight=0.5, timeout=120):
    """
    Finds the closest compounds given a Molecule.


    Parameters
    ----------
    molecule : marsi.chemistry.molecule.Molecule
        A molecule representation.
    nn_model : marsi.nearest_neighbors.model.DistributedNearestNeighbors
        A nearest neighbors model.
    fp_cut : float
        A cutoff value for fingerprint similarity.
    fpformat : str
        A valid fingerprint format.
    atoms_diff : int
        The max number of atoms that can be different (in number, not type).
    bonds_diff : int
        The max number of bonds that can be different (in number, not type).
    rings_diff : int
        The max number of rings that can be different (in number, not type).
    session : Session
        SQLAlchemy session.
    atoms_weight : float
        The weight of having matching atoms in the structural similarity
    bonds_weight : float
        The weight of having matching bonds in the structural similarity

    Returns
    -------
    pandas.DataFrame
        A data frame with the closest InChI Keys as index and the properties calculated for each hit.
    """
    assert isinstance(molecule, Molecule)

    if nn_model is None:
        query = and_(Metabolite.num_atoms >= molecule.num_atoms - atoms_diff,
                     Metabolite.num_atoms <= molecule.num_atoms + atoms_diff,
                     Metabolite.num_bonds >= molecule.num_bonds - bonds_diff,
                     Metabolite.num_bonds <= molecule.num_bonds + bonds_diff,
                     Metabolite.num_rings >= molecule.num_rings - rings_diff,
                     Metabolite.num_rings <= molecule.num_rings + rings_diff)

        nn_model = load_nearest_neighbors_model_from_db(fpformat=fpformat, custom_query=query, session=session)

    assert isinstance(nn_model, DistributedNearestNeighbors)

    neighbors = nn_model.radius_nearest_neighbors(molecule.fingerprint(fpformat), radius=1 - fp_cut)

    if molecule.inchi_key in neighbors:
        del neighbors[molecule.inchi_key]

    dataframe = DataFrame(columns=["formula", "atoms", "bonds", "tanimoto_similarity", "structural_similarity"])
    if len(neighbors) == 0:
        return dataframe

    results = []
    tasks_queue = multiprocessing.Queue()
    results_queue = multiprocessing.Queue()

    progress = ProgressBar(maxval=len(neighbors), widgets=["Processing Neighbors: ", Bar(), ETA()])

    jobs = []
    for i in range(multiprocessing.cpu_count()):
        job = DataBuilder(molecule.inchi, tasks_queue, results_queue,
                          atoms_weight, bonds_weight, timeout)
        jobs.append(job)
        job.start()

    for inchi_key, distance in six.iteritems(neighbors):
        tasks_queue.put((inchi_key, distance))

    progress.start()
    while len(results) < len(neighbors):
        try:
            res = results_queue.get(block=True, timeout=10)
        except Empty:
            continue
        else:
            results.append(res)
            if res is not None:
                dataframe.loc[res[0]] = res[1:]
            progress.update(len(results))

    progress.finish()

    for job in jobs:
        job.terminate()

    return dataframe
